# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import argparse
from pathlib import Path
from shutil import which
from typing import TYPE_CHECKING, Any, Optional, Type

if TYPE_CHECKING:
    from tsfpga.vivado.simlib_common import VivadoSimlibCommon

# Third party libraries
from vunit.ui import VUnit
from vunit.vivado.vivado import add_from_compile_order_file, create_compile_order_file
from vunit.vunit_cli import VUnitCLI

# First party libraries
import tsfpga
import tsfpga.create_vhdl_ls_config
from tsfpga.module_list import ModuleList
from tsfpga.vivado.ip_cores import VivadoIpCores
from tsfpga.vivado.simlib import VivadoSimlib


def get_arguments_cli(default_output_path: Path) -> VUnitCLI:
    """
    Get arguments for the simulation flow.

    Arguments:
        default_output_path (pathlib.Path): Will be set as default for output path arguments
            (both VUnit files and Vivado files).
    """
    cli = VUnitCLI()

    # Print default values when doing --help
    cli.parser.formatter_class = argparse.ArgumentDefaultsHelpFormatter

    # Set the supplied default value for VUnit output path. pylint: disable=protected-access
    for action in cli.parser._actions:
        if action.dest == "output_path":
            action.default = default_output_path / "vunit_out"
            break
    else:
        raise AssertionError("VUnit --output-path argument not found")

    cli.parser.add_argument(
        "--output-path-vivado",
        type=Path,
        default=default_output_path,
        help=(
            "where to place Vivado IP core and simlib files. "
            "Note that --output-path is for VUnit files"
        ),
    )

    cli.parser.add_argument(
        "--vivado-skip", action="store_true", help="skip all steps that require Vivado"
    )

    cli.parser.add_argument(
        "--ip-compile", action="store_true", help="force (re)compile of IP cores"
    )

    cli.parser.add_argument(
        "--simlib-compile", action="store_true", help="force (re)compile of Vivado simlib"
    )

    cli.parser.add_argument(
        "--vcs-minimal",
        action="store_true",
        help="compile and run only a minimal set of tests based on Version Control System history",
    )

    cli.parser.add_argument(
        "--inspect",
        action="store_true",
        help="optionally inspect some simulation result. Is only available for some modules",
    )

    return cli


class SimulationProject:
    """
    Class for setting up and handling a VUnit simulation project. Should be reusable in most cases.
    """

    def __init__(self, args: argparse.Namespace, enable_preprocessing: bool = False) -> None:
        """
        Create a VUnit project, configured according to the given arguments.

        Arguments:
            args: Command line argument namespace from ``simulate.py``.
            enable_preprocessing: If ``True``, VUnit location/check preprocessing will be enabled.
        """
        self.args = args

        self.vunit_proj = VUnit.from_args(args=args)
        self.vunit_proj.add_vhdl_builtins()
        self.vunit_proj.add_verification_components()
        self.vunit_proj.add_random()

        if enable_preprocessing:
            self.vunit_proj.enable_location_preprocessing()
            self.vunit_proj.enable_check_preprocessing()

        self.has_commercial_simulator = self.vunit_proj.get_simulator_name() != "ghdl"

    def add_modules(
        self,
        modules: ModuleList,
        modules_no_sim: Optional[ModuleList] = None,
        include_vhdl_files: bool = True,
        include_verilog_files: bool = True,
        include_systemverilog_files: bool = True,
        **setup_vunit_kwargs: Any,
    ) -> None:
        """
        Add module source files to the VUnit project.

        Arguments:
            modules: These modules will be included in the simulation project.
            modules_no_sim: These modules will be included in the simulation project,
                but their test files will not be added.
            include_vhdl_files: Optionally disable inclusion of VHDL files from
                the modules.
            include_verilog_files: Optionally disable inclusion of Verilog files from
                the modules.
            include_systemverilog_files: Optionally disable inclusion of SystemVerilog files from
                the modules.
            setup_vunit_kwargs: Further arguments that will be sent to
                :meth:`.BaseModule.setup_vunit` for each module.
                Note that this is a "kwargs" style argument; any number of named arguments can
                be sent.
        """
        modules_no_sim = ModuleList() if modules_no_sim is None else modules_no_sim

        include_unisim = not self.args.vivado_skip
        include_ip_cores = self.has_commercial_simulator and not self.args.vivado_skip

        for module in modules + modules_no_sim:
            vunit_library = self.vunit_proj.add_library(
                library_name=module.library_name, allow_duplicate=True
            )
            simulate_this_module = module not in modules_no_sim

            for hdl_file in module.get_simulation_files(
                include_tests=simulate_this_module,
                include_unisim=include_unisim,
                include_ip_cores=include_ip_cores,
                include_vhdl_files=include_vhdl_files,
                include_verilog_files=include_verilog_files,
                include_systemverilog_files=include_systemverilog_files,
            ):
                vunit_library.add_source_file(hdl_file.path)

            if simulate_this_module:
                module.setup_vunit(
                    vunit_proj=self.vunit_proj,
                    include_unisim=include_unisim,
                    include_ip_cores=include_ip_cores,
                    inspect=self.args.inspect,
                    **setup_vunit_kwargs,
                )

    def add_vivado_simlib(self) -> Optional["VivadoSimlibCommon"]:
        """
        Add Vivado simlib to the VUnit project, unless instructed not to by ``args``.
        Will compile simlib if necessary.

        Return:
            The simlib object, ``None`` if simlib was not added due to command line argument.
        """
        if self.args.vivado_skip:
            return None

        return self._add_simlib(
            output_path=self.args.output_path_vivado, force_compile=self.args.simlib_compile
        )

    def _add_simlib(self, output_path: Path, force_compile: bool) -> "VivadoSimlibCommon":
        """
        Add Vivado simlib to the VUnit project. Compile if needed.

        .. note::

            This method can be overloaded in a subclass if you want to do something more
            advanced, e.g. fetching compiled simlib from Artifactory.

        Arguments:
            output_path: Compiled simlib will be placed in sub-directory of this path.
            force_compile: Will (re)-compile simlib even if compiled artifacts exist.

        Return:
            The simlib object.
        """
        vivado_simlib = VivadoSimlib.init(output_path=output_path, vunit_proj=self.vunit_proj)
        if force_compile or vivado_simlib.compile_is_needed:
            vivado_simlib.compile()
            vivado_simlib.to_archive()

        vivado_simlib.add_to_vunit_project()

        # Code in the "vital2000" package gives GHDL errors such as "result subtype of a pure
        # function cannot have access sub-elements". Hence, relaxed rules need to be enabled when
        # using unisim.
        self.vunit_proj.set_sim_option("ghdl.elab_flags", ["-frelaxed-rules"])

        return vivado_simlib

    def add_vivado_ip_cores(
        self,
        modules: ModuleList,
        vivado_part_name: str = "xc7z020clg400-1",
        vivado_ip_core_project_class: Optional[Type[Any]] = None,
    ) -> Optional[Path]:
        """
        Generate IP cores from the modules, unless instructed not to by ``args``.
        When running with a commercial simulator they will be added to the VUnit project.

        Arguments:
            modules: IP cores from these modules will be included in the simulation project.
            vivado_part_name: Part name to be used for Vivado IP core project.
                Might have to change from default depending on what parts you have available in your
                Vivado installation.
            vivado_ip_core_project_class: Class to be used for Vivado IP core project.
                Can be left at default in most cases.

        Return:
            Path to the Vivado IP core project's ``project`` directory.
            ``None`` if Vivado IP cores were not added due to command line argument.
        """
        if self.args.vivado_skip:
            return None

        # Generate IP core simulation files. Might be used for the vhdl_ls config,
        # even if they are not added to the simulation project.
        (
            ip_core_compile_order_file,
            ip_core_vivado_project_directory,
        ) = self._generate_ip_core_files(
            modules=modules,
            output_path=self.args.output_path_vivado,
            force_generate=self.args.ip_compile,
            part_name=vivado_part_name,
            vivado_project_class=vivado_ip_core_project_class,
        )
        if self.has_commercial_simulator:
            add_from_compile_order_file(
                vunit_obj=self.vunit_proj, compile_order_file=ip_core_compile_order_file
            )

        return ip_core_vivado_project_directory

    @staticmethod
    def _generate_ip_core_files(
        modules: ModuleList,
        output_path: Path,
        force_generate: bool,
        part_name: str,
        vivado_project_class: Optional[Type[Any]] = None,
    ) -> tuple[Path, Path]:
        """
        Generate Vivado IP core files that are to be added to the VUnit project.
        Create a new project to generate files if needed.

        Arguments:
            modules: IP cores from these modules will be included.
            output_path: IP core files will be placed in sub-directory of this path.
            force_generate: Will (re)-generate files even if they exist.
            part_name: Vivado part name.
            vivado_project_class: Class to be used for Vivado IP core project.
        """
        vivado_ip_cores = VivadoIpCores(
            modules=modules,
            output_path=output_path,
            part_name=part_name,
            vivado_project_class=vivado_project_class,
        )

        if force_generate:
            vivado_ip_cores.create_vivado_project()
            vivado_project_created = True
        else:
            vivado_project_created = vivado_ip_cores.create_vivado_project_if_needed()

        if vivado_project_created:
            # If the IP core Vivado project has been (re)created we need to create
            # a new compile order file
            create_compile_order_file(
                project_file=vivado_ip_cores.vivado_project_file,
                compile_order_file=vivado_ip_cores.compile_order_file,
            )

        return vivado_ip_cores.compile_order_file, vivado_ip_cores.project_directory


def create_vhdl_ls_configuration(
    output_path: Path,
    temp_files_path: Path,
    modules: ModuleList,
    ip_core_vivado_project_directory: Optional[Path] = None,
) -> None:
    """
    Create config for vhdl_ls (https://github.com/VHDL-LS/rust_hdl).
    Granted this might no be the "correct" place for this functionality.
    But since the call is somewhat quick (~10 ms), and simulate.py is run "often" it seems an
    appropriate place in order to always have an up-to-date vhdl_ls config.

    Arguments:
        output_path: Config file will be placed here.
        temp_files_path: Some temporary files will be placed here.
        modules: These modules will be added.
        ip_core_vivado_project_directory: Vivado IP core files in this location will be added.
    """
    # Create an empty VUnit project to add files from VUnit and OSVVM library.
    # If we were to use the "real" VUnit project that we set up above instead, all the files would
    # be in the "preprocessed" folder. Hence we use an "empty" VUnit project, and add all files
    # via modules.
    vunit_proj = VUnit.from_argv(argv=["--output-path", str(temp_files_path / "vhdl_ls_vunit_out")])
    vunit_proj.add_vhdl_builtins()
    vunit_proj.add_verification_components()
    vunit_proj.add_random()
    vunit_proj.add_osvvm()

    which_vivado = which("vivado")
    vivado_location = None if which_vivado is None else Path(which_vivado)

    tsfpga.create_vhdl_ls_config.create_configuration(
        output_path=output_path,
        modules=modules,
        vunit_proj=vunit_proj,
        vivado_location=vivado_location,
        ip_core_vivado_project_directory=ip_core_vivado_project_directory,
    )



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Import before others since it modifies PYTHONPATH. pylint: disable=unused-import
import tsfpga.examples.example_pythonpath  # noqa: F401



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------
# Reusable code for Python linting.
# --------------------------------------------------------------------------------------------------

# Standard libraries
import subprocess
import sys
from pathlib import Path

# Third party libraries
import pytest

# First party libraries
import tsfpga
from tsfpga.system_utils import create_file, run_command

THIS_DIR = Path(__file__).parent.resolve()


def run_pylint(files: list[Path]) -> None:
    """
    Will raise exception if any of the ``files`` does not pass the 'pylint' lint check.
    """
    command = [sys.executable, "-m", "pylint", "--rcfile", str(THIS_DIR / "pylintrc")] + [
        str(file_path) for file_path in files
    ]

    run_command(command)


def run_black(files: list[Path]) -> None:
    """
    Will raise exception if any of the ``files`` does not pass the 'black' formatting check.
    """
    command = [sys.executable, "-m", "black", "--check", "--diff"] + [
        str(file_path) for file_path in files
    ]
    run_command(command)


def run_isort(files: list[Path], cwd: Path) -> None:
    """
    Run the 'isort' Python tool.
    Will raise exception if any of the ``files`` does not pass the check.

    Arguments:
        files (list(pathlib.Path)): The files that shall be checked.
        cwd (pathlib.Path): Path where the command shall be executed.
            The ``pyproject.toml`` file should be located here.
    """
    command = [sys.executable, "-m", "isort", "--check", "--diff"] + [
        str(file_path) for file_path in files
    ]
    run_command(command, cwd=cwd)


def run_flake8_lint(files: list[Path]) -> None:
    """
    Will raise exception if any of the ``files`` does not pass the 'flake8' lint check.
    """
    command = [sys.executable, "-m", "flake8"] + [str(file_path) for file_path in files]
    run_command(command, cwd=tsfpga.REPO_ROOT)


# pylint: disable=redefined-outer-name
@pytest.fixture
def invalid_python_code_file(tmp_path: Path) -> Path:
    ugly_code = "aa  =\ndef bb:\ncc  = 3"
    return create_file(tmp_path / "dummy_python_file.py", ugly_code)


def test_pylint_should_raise_exception_if_there_are_ugly_files(
    invalid_python_code_file: Path,
) -> None:
    with pytest.raises(subprocess.CalledProcessError):
        run_pylint(files=[invalid_python_code_file])


def test_flake8_lint_should_raise_exception_if_there_are_ugly_files(
    invalid_python_code_file: Path,
) -> None:
    with pytest.raises(subprocess.CalledProcessError):
        run_flake8_lint(files=[invalid_python_code_file])


def test_black_formatting_should_raise_exception_if_there_are_ugly_files(
    invalid_python_code_file: Path,
) -> None:
    with pytest.raises(subprocess.CalledProcessError):
        run_black(files=[invalid_python_code_file])


def test_isort_formatting_should_raise_exception_if_there_are_ugly_files(tmp_path: Path) -> None:
    file_path = create_file(
        file=tmp_path / "dummy_python_file.py",
        contents="""
import b
import a

""",
    )

    with pytest.raises(subprocess.CalledProcessError):
        run_isort(files=[file_path], cwd=tsfpga.REPO_ROOT)



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from pathlib import Path

# Third party libraries
import pytest

# First party libraries
from tsfpga.hdl_file import HdlFile


def test_file_endings():
    assert HdlFile.file_endings == (".vhd", ".vhdl", ".v", ".vh", ".sv", ".svh")


def test_file_type():
    assert HdlFile(Path("file.vhd")).type == HdlFile.Type.VHDL
    assert HdlFile(Path("file.vhdl")).type == HdlFile.Type.VHDL

    assert HdlFile(Path("file.v")).type == HdlFile.Type.VERILOG_SOURCE
    assert HdlFile(Path("file.vh")).type == HdlFile.Type.VERILOG_HEADER

    assert HdlFile(Path("file.sv")).type == HdlFile.Type.SYSTEMVERILOG_SOURCE
    assert HdlFile(Path("file.svh")).type == HdlFile.Type.SYSTEMVERILOG_HEADER


def test_unknown_file_ending_raises_exception():
    with pytest.raises(ValueError) as exception_info:
        HdlFile(Path("file.unknown"))
    assert str(exception_info.value) == "Unsupported HDL file ending: file.unknown"


def test_can_cast_to_string_without_error():
    str(HdlFile(Path("file.vhd")))



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import os
from pathlib import Path
from typing import Any, Iterator, Optional, Union

# First party libraries
from tsfpga.system_utils import file_is_in_directory


def get_git_commit(directory: Path) -> str:
    """
    Get a string describing the current git commit.
    E.g. ``"abcdef0123"`` or ``"12345678 (local changes present)"``.

    Arguments:
        directory: The directory where git commands will be run.

    Return:
        Git commit information.
    """
    git_commit = get_git_sha(directory=directory)
    if git_local_changes_present(directory=directory):
        git_commit += " (local changes present)"

    return git_commit


def get_git_sha(directory: Path) -> str:
    """
    Get a short git SHA.

    Arguments:
        directory: The directory where git commands will be run.

    Return:
        The SHA.
    """
    # Generally, eight to ten characters are more than enough to be unique within a project.
    # The linux kernel, one of the largest projects, needs 11.
    # https://git-scm.com/book/en/v2/Git-Tools-Revision-Selection#Short-SHA-1
    sha_length = 16

    if "GIT_COMMIT" in os.environ:
        return os.environ["GIT_COMMIT"][0:sha_length]

    # Import fails if "git" executable is not available, hence it can not be on top level.
    # This function should only be called if git is available.
    # pylint: disable=import-outside-toplevel
    # Third party libraries
    from git.repo import Repo

    repo = Repo(directory, search_parent_directories=True)
    git_sha = repo.head.commit.hexsha[0:sha_length]

    return git_sha


def git_local_changes_present(directory: Path) -> bool:
    """
    Check if the git repo has local changes.

    Arguments:
        directory: The directory where git commands will be run.

    Return:
        ``True`` if the repo contains changes that have been made after the last commit.
    """
    # Import fails if "git" executable is not available, hence it can not be on top level.
    # This function should only be called if git is available.
    # pylint: disable=import-outside-toplevel
    # Third party libraries
    from git.repo import Repo

    repo = Repo(directory, search_parent_directories=True)

    return repo.is_dirty()


def git_commands_are_available(directory: Path) -> bool:
    """
    True if "git" command executable is available, and ``directory`` is in a valid git repo.
    """
    try:
        # pylint: disable=import-outside-toplevel
        # Third party libraries
        from git import InvalidGitRepositoryError
        from git.repo import Repo
    except ImportError:
        return False

    try:
        Repo(directory, search_parent_directories=True)
    except InvalidGitRepositoryError:
        return False

    return True


def find_git_files(
    directory: Path,
    exclude_directories: Optional[list[Path]] = None,
    file_endings_include: Optional[Union[str, tuple[str]]] = None,
    file_endings_avoid: Optional[Union[str, tuple[str]]] = None,
) -> Iterator[Path]:
    """
    Find files that are checked in to git.

    Arguments:
        directory: Search in this directory.
        exclude_directories: Files in these directories will not be included.
        file_endings_include: Only files with these endings will be included.
        file_endings_avoid: Files with these endings will not be included.

    Return:
        The files that are available in git.
    """

    # Import fails if "git" executable is not available, hence it can not be on top level.
    # This function should only be called if git is available.
    # pylint: disable=import-outside-toplevel
    # Third party libraries
    from git.repo import Repo

    exclude_directories = (
        []
        if exclude_directories is None
        else [exclude_directory.resolve() for exclude_directory in exclude_directories]
    )

    def list_paths(root_tree: Any, path: Path) -> Iterator[Path]:
        for blob in root_tree.blobs:
            yield path / blob.name
        for tree in root_tree.trees:
            yield from list_paths(tree, path / tree.name)

    repo = Repo(directory, search_parent_directories=True)
    repo_root = Path(repo.working_dir).resolve()

    for file_path in list_paths(repo.tree(), repo_root):
        if file_endings_include is not None and not file_path.name.endswith(file_endings_include):
            continue

        if file_endings_avoid is not None and file_path.name.endswith(file_endings_avoid):
            continue

        if file_is_in_directory(file_path, exclude_directories):
            continue

        if file_is_in_directory(file_path, [directory]):
            yield file_path



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import hashlib
import json
from pathlib import Path
from typing import TYPE_CHECKING, Optional

# First party libraries
from tsfpga.system_utils import create_file, delete, read_file

# Local folder libraries
from .project import VivadoIpCoreProject

if TYPE_CHECKING:
    # First party libraries
    from tsfpga.ip_core_file import IpCoreFile
    from tsfpga.module_list import ModuleList

    # Local folder libraries
    from .project import VivadoProject


class VivadoIpCores:
    """
    Handle a list of IP core sources. Has a mechanism to detect whether a regenerate of IP files
    is needed.
    """

    project_name = "vivado_ip_project"

    def __init__(
        self,
        modules: "ModuleList",
        output_path: Path,
        part_name: str,
        vivado_project_class: Optional[type["VivadoProject"]] = None,
    ) -> None:
        """
        Arguments:
            modules: IP cores from  these modules will be included.
            output_path: The Vivado project will be placed here.
            part_name: Vivado part name to be used for the project.
            vivado_project_class: The Vivado project class that will be used for the IP core
                project. Is safe to leave at default in most cases.
        """
        self.project_directory = output_path.resolve() / self.project_name
        self._part_name = part_name

        vivado_project_class = (
            VivadoIpCoreProject if vivado_project_class is None else vivado_project_class
        )

        self._hash_file = self.project_directory / "ip_files_hash.txt"

        self._setup(modules, vivado_project_class)

    @property
    def compile_order_file(self) -> Path:
        """
        pathlib.Path: Path to the generated compile order file.
        """
        return self.project_directory / "compile_order.txt"

    @property
    def vivado_project_file(self) -> Path:
        """
        pathlib.Path: Path to the Vivado project file.
        """
        return self._vivado_project.project_file(self.project_directory)

    def create_vivado_project(self) -> None:
        """
        Create IP core Vivado project.
        """
        print(f"Creating IP core project in {self.project_directory}")
        delete(self.project_directory)
        success = self._vivado_project.create(self.project_directory)

        assert success, "Failed to create Vivado IP core project"

        self._save_hash()

    def create_vivado_project_if_needed(self) -> bool:
        """
        Create IP core Vivado project if anything has changed since last time this was run.
        If

        * List of TCL files that create IP cores,
        * and contents of these files,

        is the same then it will not create. But if anything is added or removed from the list,
        or the contents of a TCL file is changed, there will be a recreation.

        Return:
            True if Vivado project was created. False otherwise.
        """
        if self._should_create():
            self.create_vivado_project()
            return True

        return False

    def _setup(self, modules: "ModuleList", vivado_project_class: type["VivadoProject"]) -> None:
        self._vivado_project = vivado_project_class(
            name=self.project_name, modules=modules, part=self._part_name
        )

        ip_core_files = []
        for module in modules:
            # Send the same two arguments that are sent in the VivadoProject create flow
            ip_core_files += module.get_ip_core_files(generics={}, part=self._part_name)

        self._hash = self._calculate_hash(ip_core_files)

    @staticmethod
    def _calculate_hash(ip_core_files: list["IpCoreFile"]) -> str:
        """
        A string with hashes of the different IP core files.
        """
        data = ""

        def sort_by_file_name(ip_core_file: "IpCoreFile") -> str:
            return ip_core_file.path.name

        for ip_core_file in sorted(ip_core_files, key=sort_by_file_name):
            data += f"{ip_core_file.path}\n"

            if ip_core_file.variables:
                data += json.dumps(ip_core_file.variables, sort_keys=True)
                data += "\n"

            with open(ip_core_file.path, "rb") as file_handle:
                ip_hash = hashlib.md5()
                ip_hash.update(file_handle.read())
                data += f"{ip_hash.hexdigest()}\n"

        return data

    def _save_hash(self) -> None:
        create_file(self._hash_file, self._hash)

    def _should_create(self) -> bool:
        """
        Return True if a Vivado project create is needed, i.e. if anything has changed.
        """
        if not (self._hash_file.exists() and self.compile_order_file.exists()):
            return True

        return read_file(self._hash_file) != self._hash



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import sys
from pathlib import Path
from shutil import copy2, make_archive
from typing import TYPE_CHECKING

# Do PYTHONPATH insert() instead of append() to prefer any local repo checkout over any pip install
REPO_ROOT = Path(__file__).parent.parent.parent.resolve()
sys.path.insert(0, str(REPO_ROOT))

# Import before others since it modifies PYTHONPATH. pylint: disable=unused-import
import tsfpga.examples.example_pythonpath  # noqa: F401

# First party libraries
from tsfpga.build_project_list import BuildProjectList
from tsfpga.examples.build_fpga_utils import arguments, generate_register_artifacts, setup_and_run
from tsfpga.examples.example_env import TSFPGA_EXAMPLES_TEMP_DIR, get_tsfpga_example_modules
from tsfpga.system_utils import create_directory, delete

if TYPE_CHECKING:
    # First party libraries
    from tsfpga.vivado.project import VivadoProject


def main() -> None:
    """
    Main function for building FPGA projects. If you are setting up a new build flow from scratch,
    you probably want to copy and modify this function, and reuse the others.
    """
    args = arguments(default_temp_dir=TSFPGA_EXAMPLES_TEMP_DIR)
    modules = get_tsfpga_example_modules()
    projects = BuildProjectList(
        modules=modules,
        project_filters=args.project_filters,
        include_netlist_not_top_builds=args.netlist_builds,
        no_color=args.no_color,
    )

    sys.exit(
        setup_and_run(
            modules=modules,
            projects=projects,
            args=args,
            collect_artifacts_function=collect_artifacts,
        )
    )


def collect_artifacts(project: "VivadoProject", output_path: Path) -> bool:
    """
    Example of a method to collect build artifacts. Will create a zip file with the bitstream,
    hardware definition (.xsa) and register documentation.

    Arguments:
        project: Project object that has been built, and who's artifacts shall now be collected.
        output_path: Path to the build output. Artifact zip will be placed here as well.

    Return:
        True if everything went well.
    """
    version = "0.0.0"
    release_dir = create_directory(output_path / f"{project.name}-{version}", empty=True)
    print(f"Creating release in {release_dir.resolve()}.zip")

    generate_register_artifacts(modules=project.modules, output_path=release_dir / "registers")
    copy2(output_path / f"{project.name }.bit", release_dir)
    copy2(output_path / f"{project.name}.bin", release_dir)
    if (output_path / f"{project.name}.xsa").exists():
        copy2(output_path / f"{project.name}.xsa", release_dir)

    make_archive(str(release_dir), "zip", release_dir)

    # Remove folder so that only zip remains
    delete(release_dir)

    return True


if __name__ == "__main__":
    main()



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from abc import ABC, abstractmethod
from typing import Union

# Local folder libraries
from .build_result import BuildResult


class Limit(ABC):
    """
    Base class for limit checks.
    Inherit and implement the check in subclass.
    """

    def __init__(self, value: int) -> None:
        """
        Arguments:
            value: The result value shall be compared with this number.
        """
        self.value = value

    @abstractmethod
    def check(self, result_value: Union[int, float]) -> bool:
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass


class LessThan(Limit):
    """
    Limit to be used with a checker to see that a figure is less than the specified value.
    """

    def check(self, result_value: Union[int, float]) -> bool:
        return result_value < self.value

    def __str__(self) -> str:
        return f"< {self.value}"


class EqualTo(Limit):
    """
    Limit to be used with a checker to see that a figure is equal to the specified value.
    """

    def check(self, result_value: Union[int, float]) -> bool:
        return result_value == self.value

    def __str__(self) -> str:
        return str(self.value)


class BuildResultChecker(ABC):
    """
    Base class for build result checkers that check a certain build result value against a limit.

    Overload in subclass and implement the ``check`` method according to the resource you want
    to check.
    """

    def __init__(self, limit: Limit):
        """
        Arguments:
            limit: The limit that the specified resource shall be checked against.
        """
        self.limit = limit

    @abstractmethod
    def check(self, build_result: BuildResult) -> bool:
        """
        Arguments:
            build_result: Build result that shall be checked. Should come from a successful build.

        Return:
            True if check passed, false otherwise.
        """

    def _check_result_value(self, name: str, result_value: Union[int, float]) -> bool:
        if result_value is not None and self.limit.check(result_value=result_value):
            print(f"Result check passed for {name}: {result_value} ({self.limit})")
            return True

        print(f"Result check failed for {name}. Got {result_value}, expected {self.limit}.")
        return False


class MaximumLogicLevel(BuildResultChecker):
    """
    Check the maximum logic level of a build result against a limit.
    """

    name = "Maximum logic level"

    def check(self, build_result: BuildResult) -> bool:
        value = build_result.maximum_logic_level
        if value is None:
            raise ValueError("Should only call after successful synthesis")

        return self._check_result_value(name="maximum logic level", result_value=value)


class SizeChecker(BuildResultChecker):
    """
    Check a build result size value against a limit.

    Overload and set the correct ``name``, according to the names
    in the vendor utilization report.

    Note that since this is to be used by netlist builds it checks the synthesized size, not
    the implemented one, even if available.
    """

    name: str

    def check(self, build_result: BuildResult) -> bool:
        if build_result.synthesis_size is None:
            raise ValueError("Should only call after successful synthesis")

        if self.name not in build_result.synthesis_size:
            raise ValueError(
                f'Synthesis result size does not contain the requested resource: "{self.name}"'
            )

        return self._check_result_value(
            name=self.name, result_value=build_result.synthesis_size[self.name]
        )


class TotalLuts(SizeChecker):
    name = "Total LUTs"


class LogicLuts(SizeChecker):
    name = "Logic LUTs"


class LutRams(SizeChecker):
    name = "LUTRAMs"


class Srls(SizeChecker):
    name = "SRLs"


class Ffs(SizeChecker):
    name = "FFs"


class Ramb36(SizeChecker):
    name = "RAMB36"


class Ramb18(SizeChecker):
    name = "RAMB18"


class Ramb(SizeChecker):
    """
    Combined checker for RAMB36 and RAMB18 count.
    Each RAMB18 counts as half a RAMB36.
    """

    name = "RAMB"

    def check(self, build_result: BuildResult) -> bool:
        """
        Similar to super class, but takes out two result values.
        """
        ramb36_value = self._get_result_value(build_result=build_result, name=Ramb36.name)
        ramb18_value = self._get_result_value(build_result=build_result, name=Ramb18.name)

        value = ramb36_value + ramb18_value / 2

        return self._check_result_value(name=self.name, result_value=value)

    @staticmethod
    def _get_result_value(build_result: BuildResult, name: str) -> int:
        if build_result.synthesis_size is None:
            raise ValueError("Should only call after successful synthesis")

        if name not in build_result.synthesis_size:
            raise ValueError(
                f'Synthesis result size does not contain the requested resource: "{name}"'
            )

        return build_result.synthesis_size[name]


class Uram(SizeChecker):
    name = "URAM"


class DspBlocks(SizeChecker):
    """
    In Vivado pre-2020.1 the resource was called "DSP48 Blocks" in the utilization report.
    After that it is called "DSP Blocks". This class checks for both.
    """

    name = "DSP Blocks"

    def check(self, build_result: BuildResult) -> bool:
        """
        Same as super class, but checks for the legacy name as well as the current name.
        """
        if build_result.synthesis_size is None:
            raise ValueError("Should only call after successful synthesis")

        legacy_name = "DSP48 Blocks"

        if legacy_name in build_result.synthesis_size:
            value = build_result.synthesis_size[legacy_name]
        else:
            value = build_result.synthesis_size[self.name]

        return self._check_result_value(name=self.name, result_value=value)



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional

# Third party libraries
import rtoml

# First party libraries
from tsfpga.vivado.ip_cores import VivadoIpCores

if TYPE_CHECKING:
    # Local folder libraries
    from .module_list import ModuleList


def create_configuration(
    output_path: Path,
    modules: Optional["ModuleList"] = None,
    vunit_proj: Optional[Any] = None,
    vivado_location: Optional[Path] = None,
    ip_core_vivado_project_directory: Optional[Path] = None,
) -> None:
    """
    Create a configuration file (vhdl_ls.toml) for the rust_hdl VHDL Language Server.

    Can be used with modules and an "empty" VUnit project, or with a complete VUnit
    project with all user files added.

    Execution of this function takes roughly 12 ms for a large project (62 modules and a
    VUnit project).

    Arguments:
        output_path: vhdl_ls.toml file will be placed in this folder.
        modules: A list of Module objects.
        vunit_proj: A VUnit project.
        vivado_location: Vivado binary path. Will add unisim from this Vivado installation.
        ip_core_vivado_project_directory: Path to a Vivado project that contains
            generated "simulation" and "synthesis" files of IP cores
            (the "generate_target" TCL command). See simulate.py for an example of using this.
    """
    toml_data: dict[str, dict[str, dict[str, list[str]]]] = dict(libraries={})

    if modules is not None:
        for module in modules:
            vhd_file_wildcard = module.path.resolve() / "**" / "*.vhd"
            toml_data["libraries"][module.library_name] = dict(files=[str(vhd_file_wildcard)])

    if vunit_proj is not None:
        for source_file in vunit_proj.get_compile_order():
            if source_file.library.name not in toml_data["libraries"]:
                toml_data["libraries"][source_file.library.name] = dict(files=[])
            toml_data["libraries"][source_file.library.name]["files"].append(
                str(Path(source_file.name).resolve())
            )

    if vivado_location is not None:
        vcomponents_package = (
            vivado_location.parent.parent
            / "data"
            / "vhdl"
            / "src"
            / "unisims"
            / "unisim_retarget_VCOMP.vhd"
        )
        if not vcomponents_package.exists():
            raise FileNotFoundError(f"Could not find unisim file: {vcomponents_package}")

        toml_data["libraries"]["unisim"] = dict(files=[str(vcomponents_package.resolve())])

    if ip_core_vivado_project_directory is not None:
        ip_core_vivado_project_directory = ip_core_vivado_project_directory.resolve()
        toml_data["libraries"]["xil_defaultlib"] = dict(files=[])

        # Vivado 2020.2+ (?) seems to place the files in "gen"
        ip_gen_dir = (
            ip_core_vivado_project_directory / f"{VivadoIpCores.project_name}.gen" / "sources_1"
        )
        if ip_gen_dir.exists():
            vhd_file_wildcard = ip_gen_dir / "ip" / "**" / "*.vhd"
            toml_data["libraries"]["xil_defaultlib"]["files"].append(str(vhd_file_wildcard))

    rtoml.dump(obj=toml_data, file=output_path / "vhdl_ls.toml", pretty=True)



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------


def to_binary_string(value: int, result_width: int) -> str:
    """
    Convert unsigned integer value to a zero-padded string of 1's and 0's.
    Most significant bit is the first (left-most) character in the string.

    For example, value 37, width 6, returns ``100101``.
    For example, value 37, width 8, returns ``00100101``.

    Arguments:
        value: The value to be converted.
        result_width_bits: The supplied ``value`` will be interpreted as an unsigned value with
            this many bits.
            The result string will contain this many bit characters.
    """
    _check_unsigned_range(value=value, width=result_width)

    # Cast to binary, pad with zeros on the left.
    formatting_string = f"{{:0{result_width}b}}"
    padded_binary_string = formatting_string.format(value)
    assert len(padded_binary_string) == result_width

    return padded_binary_string


def to_binary_nibble_string(value: int, result_width_bits: int) -> str:
    """
    Convert unsigned integer value to a zero-padded string of 1's and 0's, with each nibble (4 bits)
    separated by "_".
    Most significant bit is the first (left-most) character in the string.

    For example, value 37, width 6, returns ``10_0101``.
    For example, value 37, width 8, returns ``0010_0101``.

    Arguments:
        value: The value to be converted.
        result_width_bits: The supplied ``value`` will be interpreted as an unsigned value with
            this many bits.
            The result string will contain this many bit characters, plus separators.
    """
    _check_unsigned_range(value=value, width=result_width_bits)

    result_width_nibbles = (result_width_bits + 4 - 1) // 4
    num_separators = result_width_nibbles - 1
    result_width = result_width_bits + num_separators

    # Cast to binary, pad with zeros on the left, separate every fourth character.
    formatting_string = f"{{:0{result_width}_b}}"
    result = formatting_string.format(value)
    assert len(result) == result_width

    return result


def to_hex_string(value: int, result_width_bits: int) -> str:
    """
    Convert unsigned integer value to a zero-padded string of 01ABCDEF.
    Most significant bit is the first (left-most) character in the string.

    For example, value 60, width 6, returns ``3C``.
    For example, value 60, width 9, returns ``03C``.

    Arguments:
        value: The value to be converted.
        result_width_bits: The supplied ``value`` will be interpreted as an unsigned value with
            this many bits.
            The result string will contain enough hex characters to represent this many
            bits (rounded up).
    """
    _check_unsigned_range(value=value, width=result_width_bits)

    result_width_nibbles = (result_width_bits + 4 - 1) // 4

    # Cast to hex, pad with zeros on the left.
    formatting_string = f"{{:0{result_width_nibbles}X}}"
    result = formatting_string.format(value)
    assert len(result) == result_width_nibbles

    return result


def to_hex_byte_string(value: int, result_width_bits: int) -> str:
    """
    Convert unsigned integer value to a zero-padded string of 01ABCDEF, with each byte
    (8 bits, 2 result characters) separated by "_".
    Most significant bit is the first (left-most) character in the string.

    For example, value 60, width 6, returns ``3C``.
    For example, value 60, width 9, returns ``0_3C``.

    Arguments:
        value: The value to be converted.
        result_width_bits: The supplied ``value`` will be interpreted as an unsigned value with
            this many bits.
            The result string will contain enough hex characters to represent this many
            bits (rounded up), plus separators.
    """
    hex_string = to_hex_string(value=value, result_width_bits=result_width_bits)

    result_width_nibbles = (result_width_bits + 4 - 1) // 4
    assert len(hex_string) % result_width_nibbles == 0, hex_string

    byte_strings = []
    if result_width_nibbles % 2 == 1:
        byte_strings.append(hex_string[0])
    for i in range(result_width_nibbles % 2, result_width_nibbles, 2):
        byte_strings.append(hex_string[i : i + 2])

    result = "_".join(byte_strings)

    result_width_bytes = (result_width_bits + 8 - 1) // 8
    num_separators = result_width_bytes - 1
    result_width = result_width_nibbles + num_separators
    assert len(result) == result_width, result

    return result


def _check_unsigned_range(value: int, width: int) -> None:
    if width < 1:
        raise ValueError(f'Invalid result width "{width}".')

    if not 0 <= value < 2**width:
        raise ValueError(f'Value "{value}" out of {width}-bit range.')



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from pathlib import Path


class BuildStepTclHook:
    """
    Represent a TCL file that shall be used as hook in one of the build steps.
    """

    def __init__(self, tcl_file: Path, hook_step: str) -> None:
        """
        Arguments:
            tcl_file: Path to a TCL file.
            hook_step: Name of a build step, e.g. ``STEPS.ROUTE_DESIGN.TCL.PRE``.
               See
               https://docs.xilinx.com/r/en-US/ug894-vivado-tcl-scripting/Defining-Tcl-Hook-Scripts
               for a list of the available build steps in AMD Vivado.
        """
        self.tcl_file = tcl_file
        self.hook_step = hook_step

    @property
    def step_is_synth(self) -> bool:
        """
        True if the build step is in synthesis. False otherwise.
        """
        return "synth" in self.hook_step.lower()

    def __str__(self) -> str:
        result = str(self.__class__.__name__) + ":"
        return result + f"{self.hook_step}:{self.tcl_file}"



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

"""
Configuration file for the Sphinx documentation builder.
For building documentation of the example modules.
"""

project = "tsfpga example modules"
copyright = "Lukas Vik"
author = "Lukas Vik"

extensions = [
    "sphinx_rtd_theme",
    "symbolator_sphinx",
]

symbolator_output_format = "png"

html_theme = "sphinx_rtd_theme"



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import re
from pathlib import Path
from typing import Optional

# First party libraries
from tsfpga.system_utils import read_file

SEPARATOR_LINE_LENGTH = 100
VHDL_COMMENT_SEPARATOR = "-- " + ("-" * (SEPARATOR_LINE_LENGTH - 3))


class VhdlFileDocumentation:
    """
    Methods to extract documentation from a VHDL source file.
    """

    def __init__(self, vhd_file_path: Path) -> None:
        """
        Arguments:
            vhd_file_path: Path to the VHDL file which shall be documented.
        """
        self._vhd_file_path = vhd_file_path

    def get_header_rst(self) -> Optional[str]:
        """
        Get the contents of the VHDL file's header. This means everything that is in the comment
        block at the start of the file, after the copyright notice.

        Return:
            File header content.
        """
        file_contents = read_file(self._vhd_file_path)

        documentation_header_regexp = re.compile(
            VHDL_COMMENT_SEPARATOR
            + r"\n(.+?)\n"
            + VHDL_COMMENT_SEPARATOR
            + r"\n(.+?)\n"
            + VHDL_COMMENT_SEPARATOR
            + r"\n\n",
            re.DOTALL,
        )
        match = documentation_header_regexp.search(file_contents)
        if match is None:
            return None

        # The first group will match the copyright header. Second group is documentation.
        lines = match.group(2).split("\n")
        text = ""
        for line in lines:
            if line == "--":
                text += "\n"
            else:
                # Remove initial "-- " from comments
                text += f"{line[3:]}\n"

        return text

    def get_symbolator_component(  # pylint: disable=too-many-locals,too-many-statements
        self,
    ) -> Optional[str]:
        """
        Return a string with a ``component`` declaration equivalent to the ``entity`` declaration
        within the file. (We use entity's but symbolator requires component's).

        Default values and range declarations on ports are removed since symbolator does not seem
        to handle them.

        This implementation uses some regular expressions to find the generics and ports and modify
        them.
        The use of regular expressions makes it somewhat simple but also limited.
        Comments in strange places, specifically the string ``port (`` in a comment will make the
        mechanism fail.
        Also an entity with generics but no ports will be falsely interpreted as only ports.

        These known limitations do not pose any known practical problem and are hence considered
        worth it in order to keep the implementation simple.
        The real solution would be to fix upstream in symbolator and hdlparse.

        Return:
            VHDL ``component`` declaration.
            ``None`` if file is a package, and hence contains no ``entity``.
            ``None`` if no ``entity`` is found in the file.
        """
        if self._vhd_file_path.name.endswith("_pkg.vhd"):
            # File is a package, hence contains no entity
            return None

        vhdl = read_file(self._vhd_file_path)

        def replace_comment(match):  # type: ignore
            match_group = match.group(1)
            if (
                match_group
                and match_group.startswith("# {{")
                and match_group.endswith("}}")
                and ";" not in match_group
            ):
                # This is a valid symbolator comment, leave it as is.
                # I.e. "--# {{}}" or "--# {{some text}}
                return f"--{match_group}"

            # Strip comment
            return ""

        # Remove comments so that the remaining VHDL is easier to parse.
        vhdl = re.sub(pattern=r"--(.*)$", repl=replace_comment, string=vhdl, flags=re.MULTILINE)

        # Strip trailing whitespace and empty lines.
        vhdl = re.sub(pattern=r"\s*$", repl="", string=vhdl, flags=re.MULTILINE)

        # Split out the entity declaration from the VHDL file.
        entity_name = self._vhd_file_path.stem
        entity_regexp = re.compile(
            rf"entity\s+{entity_name}\s+is\s+"
            # Match all the code for generics and ports.
            # Is non-greedy, so it will only match up until the "end" declaration below.
            r"(.*?)"
            # Shall be able to handle
            #   end entity;
            #   end entity <name>;
            #   end <name>;
            #   end;
            # with different whitespace around.
            rf"\s*end(\s+entity|\s+entity\s+{entity_name}|\s+{entity_name}|)\s*;",
            re.IGNORECASE | re.DOTALL,
        )

        match = entity_regexp.search(vhdl)
        if match is None or not match.group(1):
            print(f"Found no entity in {self._vhd_file_path}")
            return None

        ports_and_generics = match.group(1)

        # Remove attribute lines within the entity declaration.
        ports_and_generics = re.sub(
            pattern=r"^\s*attribute\s+.+", repl="", string=ports_and_generics, flags=re.MULTILINE
        )

        # Slit out the generic part and the port part from the entity declaration.
        ports_and_generics_regexp = re.compile(
            # Match all the code for generics and ports.
            # Is non-greedy, so it will only match up until the "end" declaration below.
            # Generic block optionally
            r"\s*(.+?)?\s*(\)\s*;\s*)?"
            # Port block
            r"port\s*\(\s*(.+?)\s*"
            #
            r"\)\s*;\s*$",
            re.IGNORECASE | re.DOTALL,
        )

        match = ports_and_generics_regexp.search(ports_and_generics)
        if match is None:
            print(f"Found no ports or generics in {self._vhd_file_path}")
            return None

        if match.group(2) and match.group(3):
            # Entity declaration contains both generics and ports
            generics = match.group(1)
            assert generics.lower().startswith("generic")

            strip_generics_open = re.compile(r"generic\s*\(", re.IGNORECASE)
            generics = strip_generics_open.sub(repl="", string=generics)
        else:
            # Only one match, which we assume is ports (generics only is not supported)
            generics = None

        ports = match.group(3)

        # Remove default values.
        # Symbolator stops parsing if it encounters vector default values (others => ...).
        default_value_regexp = re.compile(r"\s*:=.+$", re.IGNORECASE | re.DOTALL)

        # Remove any vector range declarations in port/generic list.
        # The lines become too long so they don't fit in the image.
        vector_regexp = re.compile(r"\(.*$", re.IGNORECASE | re.DOTALL)

        def clean_up_declarations(declarations: str) -> str:
            clean_declarations = []

            # Split the list of declarations string into individual.
            # Note that this fails if there are any ";" in comments, so its import that we
            # strip comments before this.
            for declaration in declarations.split(";"):
                cleaned = default_value_regexp.sub(repl="", string=declaration)
                cleaned = vector_regexp.sub(repl="", string=cleaned)

                clean_declarations.append(cleaned)

            return ";".join(clean_declarations)

        if generics:
            generics = clean_up_declarations(generics)

        ports = clean_up_declarations(ports)

        if generics:
            generics_code = f"  generic ({generics}\n  );\n"
        else:
            generics_code = ""

        ports_code = f"  port (\n    {ports}\n  );"

        component = f"""\
component {entity_name} is
{generics_code}{ports_code}
end component;"""

        return component



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from pathlib import Path
from unittest.mock import ANY, MagicMock, patch

# Third party libraries
import pytest

# First party libraries
from tsfpga.module import BaseModule, get_modules
from tsfpga.system_utils import create_directory, create_file


def test_add_vunit_config_name():
    module = BaseModule(path=Path(), library_name="")

    test = MagicMock()
    pre_config = MagicMock()
    post_check = MagicMock()

    module.add_vunit_config(test=test, pre_config=pre_config, post_check=post_check)
    test.add_config.assert_called_once_with(
        name="test", generics={}, pre_config=pre_config, post_check=post_check
    )
    test.reset_mock()

    module.add_vunit_config(test=test, name="apa")
    test.add_config.assert_called_once_with(
        name="apa", generics={}, pre_config=None, post_check=None
    )
    test.reset_mock()

    module.add_vunit_config(test=test, generics=dict(apa="hest", foo="bar"))
    test.add_config.assert_called_once_with(
        name="apa_hest.foo_bar",
        generics=dict(apa="hest", foo="bar"),
        pre_config=None,
        post_check=None,
    )
    test.reset_mock()

    module.add_vunit_config(test=test, name="zebra", generics=dict(apa="hest", foo="bar"))
    test.add_config.assert_called_once_with(
        name="zebra.apa_hest.foo_bar",
        generics=dict(apa="hest", foo="bar"),
        pre_config=None,
        post_check=None,
    )


def test_add_vunit_config_random_seed():
    module = BaseModule(path=Path(), library_name="")
    test = MagicMock()

    # No seed at all
    module.add_vunit_config(test=test)
    assert "generics" not in test.add_config.call_args

    module.add_vunit_config(test=test, set_random_seed=False)
    assert "generics" not in test.add_config.call_args

    # No seed, with generics set
    module.add_vunit_config(test=test, generics={"apa": "whatever"})
    assert "seed" not in test.add_config.call_args.kwargs["generics"]

    # Static seed
    module.add_vunit_config(test=test, set_random_seed=0)
    assert isinstance(test.add_config.call_args.kwargs["generics"]["seed"], int)
    assert test.add_config.call_args.kwargs["generics"]["seed"] == 0

    module.add_vunit_config(test=test, set_random_seed=123)
    assert isinstance(test.add_config.call_args.kwargs["generics"]["seed"], int)
    assert test.add_config.call_args.kwargs["generics"]["seed"] == 123

    # Use random seed
    module.add_vunit_config(test=test, set_random_seed=True)
    assert isinstance(test.add_config.call_args.kwargs["generics"]["seed"], int)
    assert test.add_config.call_args.kwargs["generics"]["seed"] >= 0

    # Setting explicit value should still work
    module.add_vunit_config(test=test, generics={"seed": 711})
    assert test.add_config.call_args.kwargs["generics"]["seed"] == 711

    # If a value is already set it will be overwritten
    module.add_vunit_config(test=test, generics={"seed": -5}, set_random_seed=True)
    assert test.add_config.call_args.kwargs["generics"]["seed"] != -5


def test_file_list_filtering(tmp_path):
    module_name = "zebra"
    path = tmp_path / module_name

    create_directory(path / "folder_should_not_be_included")
    create_file(path / "should_not_be_included.apa")

    synth_files = {
        create_file(path / "syn.v"),
        create_file(path / "rtl" / "syn.v"),
        create_file(path / "src" / "syn.vhd"),
        create_file(path / "hdl" / "rtl" / "syn.vhdl"),
        create_file(path / "hdl" / "package" / "syn.vhd"),
    }

    test_files = {
        create_file(path / "test" / "test.v"),
        create_file(path / "rtl" / "tb" / "test.vhd"),
    }

    sim_files = {create_file(path / "sim" / "sim.vhd")}

    my_module = BaseModule(path=path, library_name="zebra")

    files = {file.path for file in my_module.get_synthesis_files()}
    assert files == synth_files

    files = {file.path for file in my_module.get_simulation_files()}
    assert files == synth_files | test_files | sim_files

    files = {file.path for file in my_module.get_simulation_files(include_tests=False)}
    assert files == synth_files | sim_files

    files = {file.path for file in my_module.get_simulation_files(files_include=synth_files)}
    assert files == synth_files

    files = {file.path for file in my_module.get_simulation_files(files_avoid=synth_files)}
    assert files == test_files | sim_files


def test_get_synthesis_files_calls_get_simulation_files_with_correct_arguments():
    module = BaseModule(path=Path(), library_name="")
    with patch("tsfpga.module.BaseModule.get_synthesis_files") as get_synthesis_files:
        module.get_simulation_files(
            files_include=True,
            files_avoid=False,
            apa=123,
            include_vhdl_files=1,
            include_verilog_files=2,
            include_systemverilog_files=3,
        )
        get_synthesis_files.assert_called_once_with(
            files_include=True,
            files_avoid=False,
            apa=123,
            include_vhdl_files=1,
            include_verilog_files=2,
            include_systemverilog_files=3,
        )


def test_get_vhdl_files(tmp_path):
    paths = {
        create_file(tmp_path / "apa.vhdl"),
        create_file(tmp_path / "apa.vhd"),
    }
    create_file(tmp_path / "apa.v")
    create_file(tmp_path / "apa.vh")
    create_file(tmp_path / "apa.sv")
    create_file(tmp_path / "apa.svh")

    got_hdl_files = BaseModule(path=tmp_path, library_name="").get_synthesis_files(
        include_verilog_files=False, include_systemverilog_files=False
    )
    assert {hdl_file.path for hdl_file in got_hdl_files} == paths


def test_get_verilog_files(tmp_path):
    paths = {create_file(tmp_path / "apa.v"), create_file(tmp_path / "apa.vh")}
    create_file(tmp_path / "apa.vhdl")
    create_file(tmp_path / "apa.vhd")
    create_file(tmp_path / "apa.sv")
    create_file(tmp_path / "apa.svh")

    got_hdl_files = BaseModule(path=tmp_path, library_name="").get_simulation_files(
        include_vhdl_files=False, include_systemverilog_files=False
    )
    assert {hdl_file.path for hdl_file in got_hdl_files} == paths


def test_get_systemverilog_files(tmp_path):
    paths = {create_file(tmp_path / "apa.sv"), create_file(tmp_path / "apa.svh")}
    create_file(tmp_path / "apa.vhdl")
    create_file(tmp_path / "apa.vhd")
    create_file(tmp_path / "apa.v")
    create_file(tmp_path / "apa.vh")

    got_hdl_files = BaseModule(path=tmp_path, library_name="").get_documentation_files(
        include_vhdl_files=False, include_verilog_files=False
    )
    assert {hdl_file.path for hdl_file in got_hdl_files} == paths


def test_get_documentation_files(tmp_path):
    module_name = "zebra"
    path = tmp_path / module_name

    synth_files = {
        create_file(path / "rtl" / "syn.v"),
        create_file(path / "src" / "syn.vhd"),
    }

    # Test files
    create_file(path / "test" / "test.v")
    create_file(path / "rtl" / "tb" / "test.vhd")

    sim_files = {create_file(path / "sim" / "sim.vhd")}

    module = BaseModule(path=path, library_name="zebra")

    # Should include everything except test files
    files = {file.path for file in module.get_documentation_files()}
    assert files == synth_files | sim_files


def test_scoped_constraints(tmp_path):
    module_path = tmp_path / "apa"
    create_file(module_path / "src" / "hest.vhd")
    create_file(module_path / "scoped_constraints" / "hest.tcl")

    my_module = BaseModule(module_path, "apa")
    scoped_constraints = my_module.get_scoped_constraints()
    assert len(scoped_constraints) == 1
    assert scoped_constraints[0].ref == "hest"


def test_scoped_constraint_entity_not_existing_should_raise_error(tmp_path):
    module_path = tmp_path / "apa"
    create_file(module_path / "scoped_constraints" / "hest.tcl")

    module = BaseModule(module_path, "apa")
    with pytest.raises(FileNotFoundError) as exception_info:
        module.get_scoped_constraints()
    assert str(exception_info.value).startswith("Could not find a matching entity file")


def test_can_cast_to_string_without_error():
    str(BaseModule(Path("dummy"), "dummy"))


def test_test_case_name():
    assert (
        BaseModule.test_case_name(generics=dict(apa=3, hest_zebra="foo")) == "apa_3.hest_zebra_foo"
    )
    assert (
        BaseModule.test_case_name(name="foo", generics=dict(apa=3, hest_zebra="bar"))
        == "foo.apa_3.hest_zebra_bar"
    )


def test_getting_registers_calls_registers_hook(tmp_path):
    with patch("tsfpga.module.from_toml", autospec=True) as from_toml, patch(
        "tsfpga.module.BaseModule.registers_hook", autospec=True
    ) as registers_hook:
        create_file(tmp_path / "a" / "regs_a.toml")
        module = BaseModule(path=tmp_path / "a", library_name="a")
        registers = module.registers

        # TOML file exists so register creation from TOML should run
        from_toml.assert_called_once()
        registers_hook.assert_called_once()
        assert registers is not None

    with patch("tsfpga.module.from_toml", autospec=True) as from_toml, patch(
        "tsfpga.module.BaseModule.registers_hook", autospec=True
    ) as registers_hook:
        module = BaseModule(path=tmp_path / "b", library_name="b")
        registers = module.registers

        # TOML file does not exist, so register creation from TOML should not run
        from_toml.assert_not_called()
        # Register hook shall still run however
        registers_hook.assert_called_once()
        assert registers is None


def test_creating_synthesis_files_does_not_create_simulation_files(tmp_path):
    create_file(tmp_path / "a" / "regs_a.toml", "register.apa.mode = 'r_w'")
    module = BaseModule(path=tmp_path / "a", library_name="a")

    synthesis_file = module.register_synthesis_folder / "a_regs_pkg.vhd"
    simulation_file = module.register_simulation_folder / "a_register_read_write_pkg.vhd"

    module.get_synthesis_files()
    assert synthesis_file.exists()
    assert not simulation_file.exists()
    assert not module.register_simulation_folder.exists()

    module.get_simulation_files()
    assert simulation_file.exists()


def test_old_register_package_should_be_deleted(tmp_path):
    create_file(tmp_path / "a" / "regs_a.toml", "register.apa.mode = 'r_w'")
    regs_pkg = create_file(tmp_path / "a" / "a_regs_pkg.vhd")

    module = BaseModule(path=tmp_path / "a", library_name="a")
    module.get_synthesis_files()

    assert not regs_pkg.exists()


@pytest.fixture
def get_modules_test(tmp_path):
    class GetModulesTest:
        def __init__(self):
            create_directory(tmp_path / "a")
            create_directory(tmp_path / "b")
            create_directory(tmp_path / "c")

            self.modules_folder = tmp_path
            self.modules_folders = [self.modules_folder]

    return GetModulesTest()


# False positive for pytest fixtures
# pylint: disable=redefined-outer-name


def test_name_filtering_include(get_modules_test):
    modules = get_modules(get_modules_test.modules_folders, names_include=["a", "b"])
    assert set(module.name for module in modules) == set(["a", "b"])


def test_name_filtering_avoid(get_modules_test):
    modules = get_modules(get_modules_test.modules_folders, names_avoid=["a", "b"])
    assert set(module.name for module in modules) == set(["c"])


def test_name_filtering_include_and_avoid(get_modules_test):
    modules = get_modules(
        get_modules_test.modules_folders, names_include=["a", "c"], names_avoid=["b", "c"]
    )
    assert set(module.name for module in modules) == set(["a"])


def test_library_name_does_not_have_lib_suffix(get_modules_test):
    modules = get_modules(get_modules_test.modules_folders)
    assert set(module.library_name for module in modules) == set(["a", "b", "c"])


def test_library_name_has_lib_suffix(get_modules_test):
    modules = get_modules(get_modules_test.modules_folders, library_name_has_lib_suffix=True)
    assert set(module.library_name for module in modules) == set(["a_lib", "b_lib", "c_lib"])


def test_stray_file_can_exist_in_modules_folder_without_error(get_modules_test):
    create_file(get_modules_test.modules_folder / "text_file.txt")
    modules = get_modules(get_modules_test.modules_folders)
    assert len(modules) == 3


def test_local_override_of_module_type(get_modules_test):
    module_file_content = """
from tsfpga.module import BaseModule

class Module(BaseModule):
    def id(self):
        return """

    create_file(get_modules_test.modules_folder / "a" / "module_a.py", module_file_content + '"a"')
    create_file(get_modules_test.modules_folder / "b" / "module_b.py", module_file_content + '"b"')

    modules = get_modules(get_modules_test.modules_folders)

    assert len(modules) == 3
    for module in modules:
        if module.name == "a":
            assert module.id() == "a"
        elif module.name == "b":
            assert module.id() == "b"
        elif module.name == "c":
            assert isinstance(module, BaseModule)
        else:
            assert False


@patch("tsfpga.module.from_toml", autospec=True)
@patch("tsfpga.module.VhdlRegisterPackageGenerator.create", autospec=True)
@patch("tsfpga.module.VhdlRecordPackageGenerator.create", autospec=True)
@patch("tsfpga.module.VhdlSimulationReadWritePackageGenerator.create", autospec=True)
@patch("tsfpga.module.VhdlSimulationWaitUntilPackageGenerator.create", autospec=True)
@patch("tsfpga.module.VhdlAxiLiteWrapperGenerator.create", autospec=True)
def test_register_toml_file_parsed_only_once_when_getting_synthesis_files(
    _, __, ___, ____, _____, from_toml, tmp_path
):
    toml_file = create_file(tmp_path / "a" / "regs_a.toml")

    module = get_modules([tmp_path]).get("a")
    module.get_synthesis_files()
    module.get_synthesis_files()

    from_toml.assert_called_once_with("a", toml_file, ANY)


@patch("tsfpga.module.from_toml", autospec=True)
@patch("tsfpga.module.VhdlRegisterPackageGenerator.create", autospec=True)
@patch("tsfpga.module.VhdlRecordPackageGenerator.create", autospec=True)
@patch("tsfpga.module.VhdlAxiLiteWrapperGenerator.create", autospec=True)
@patch("tsfpga.module.VhdlSimulationReadWritePackageGenerator.create", autospec=True)
@patch("tsfpga.module.VhdlSimulationCheckPackageGenerator.create", autospec=True)
@patch("tsfpga.module.VhdlSimulationWaitUntilPackageGenerator.create", autospec=True)
def test_register_toml_file_parsed_only_once_when_getting_simulation_files(
    _, __, ___, ____, _____, ______, from_toml, tmp_path
):
    toml_file = create_file(tmp_path / "a" / "regs_a.toml")

    module = get_modules([tmp_path]).get("a")
    module.get_simulation_files()
    module.get_simulation_files()

    from_toml.assert_called_once_with("a", toml_file, ANY)


@patch("tsfpga.module.from_toml", autospec=True)
@patch("tsfpga.module.VhdlRegisterPackageGenerator.create", autospec=True)
@patch("tsfpga.module.VhdlRecordPackageGenerator.create", autospec=True)
@patch("tsfpga.module.VhdlAxiLiteWrapperGenerator.create", autospec=True)
@patch("tsfpga.module.VhdlSimulationReadWritePackageGenerator.create", autospec=True)
@patch("tsfpga.module.VhdlSimulationCheckPackageGenerator.create", autospec=True)
@patch("tsfpga.module.VhdlSimulationWaitUntilPackageGenerator.create", autospec=True)
def test_register_toml_file_parsed_only_once_when_getting_mixed_files(
    _, __, ___, ____, _____, ______, from_toml, tmp_path
):
    toml_file = create_file(tmp_path / "a" / "regs_a.toml")

    module = get_modules([tmp_path]).get("a")
    module.get_synthesis_files()
    module.get_simulation_files()

    from_toml.assert_called_once_with("a", toml_file, ANY)



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import fnmatch
import time
from pathlib import Path
from threading import Lock
from typing import TYPE_CHECKING, Any, Callable, Iterable, Optional, Union

# Third party libraries
from vunit.color_printer import COLOR_PRINTER, NO_COLOR_PRINTER
from vunit.test.list import TestList
from vunit.test.report import TestReport, TestResult
from vunit.test.runner import TestRunner

# First party libraries
from tsfpga.system_utils import create_directory, read_last_lines_of_file

if TYPE_CHECKING:
    # Local folder libraries
    from .module_list import ModuleList
    from .vivado import build_result
    from .vivado.project import VivadoProject


class BuildProjectList:
    """
    Interface to handle a list of FPGA build projects.
    Enables building many projects in parallel.
    """

    def __init__(
        self,
        modules: "ModuleList",
        project_filters: list[str],
        include_netlist_not_top_builds: bool = False,
        no_color: bool = False,
    ):
        """
        Arguments:
            modules: Module objects that can define build projects.
            project_filters: Project name filters. Can use wildcards (*). Leave empty for all.
            include_netlist_not_top_builds: Set True to get only netlist builds,
                instead of only top level builds.
            no_color: Disable color in printouts.
        """
        self._modules = modules
        self._no_color = no_color

        self.projects = list(
            self._iterate_projects(
                project_filters=project_filters,
                include_netlist_not_top_builds=include_netlist_not_top_builds,
            )
        )

        if not self.projects:
            print(f"No projects matched this filter: {' '.join(project_filters)}")

    def __str__(self) -> str:
        """
        Returns a string with a description list of the projects.

        Will print some information about each project (name, generics, part, ...) so can become
        long if there are many projects present.
        An alternative in that case would be :meth:`.get_short_str`.
        """
        result = "\n".join([str(project) for project in self.projects])
        result += "\n"
        result += "\n"
        result += f"Listed {len(self.projects)} builds"

        return result

    def get_short_str(self) -> str:
        """
        Returns a short string with a description list of the projects.

        This is an alternative function that is more compact than ``__str__``.
        """
        result = "\n".join([project.name for project in self.projects])
        result += "\n"
        result += f"Listed {len(self.projects)} builds"

        return result

    def create(self, projects_path: Path, num_parallel_builds: int, **kwargs: Any) -> bool:
        """
        Create build project on disk for all the projects in the list.

        Arguments:
            projects_path: The projects will be placed here.
            num_parallel_builds: The number of projects that will be created in parallel.
            kwargs: Other arguments as accepted by :meth:`.VivadoProject.create`.

                .. Note::
                    Argument ``project_path`` can not be set, it is set by this class
                    based on the ``project_paths`` argument to this function.

        Return:
            True if everything went well.
        """
        build_wrappers = []
        for project in self.projects:
            build_wrapper = BuildProjectCreateWrapper(project, **kwargs)
            build_wrappers.append(build_wrapper)

        return self._run_build_wrappers(
            projects_path=projects_path,
            build_wrappers=build_wrappers,
            num_parallel_builds=num_parallel_builds,
        )

    def create_unless_exists(
        self, projects_path: Path, num_parallel_builds: int, **kwargs: Any
    ) -> bool:
        """
        Create build project for all the projects in the list, unless the project already
        exists.

        Arguments:
            projects_path: The projects will be placed here.
            num_parallel_builds: The number of projects that will be created in parallel.
            kwargs: Other arguments as accepted by :meth:`.VivadoProject.create`.

                .. Note::
                    Argument ``project_path`` can not be set, it is set by this class
                    based on the ``project_paths`` argument to this function.

        Return:
            True if everything went well.
        """
        build_wrappers = []
        for project in self.projects:
            if not (projects_path / project.name / "project").exists():
                build_wrapper = BuildProjectCreateWrapper(project, **kwargs)
                build_wrappers.append(build_wrapper)

        if not build_wrappers:
            # Return straight away if no projects need to be created. To avoid extra
            # "No tests were run!" printout from creation step that is very misleading.
            return True

        return self._run_build_wrappers(
            projects_path=projects_path,
            build_wrappers=build_wrappers,
            num_parallel_builds=num_parallel_builds,
        )

    def build(
        self,
        projects_path: Path,
        num_parallel_builds: int,
        num_threads_per_build: int,
        output_path: Optional[Path] = None,
        collect_artifacts: Optional[Callable[["VivadoProject", Path], bool]] = None,
        **kwargs: Any,
    ) -> bool:
        """
        Build all the projects in the list.

        Arguments:
            projects_path: The projects are placed here.
            num_parallel_builds: The number of projects that will be built in parallel.
            num_threads_per_build: The number threads that will be used for each
                parallel build process.
            output_path: Where the artifacts should be placed.
                Will default to within the ``projects_path`` if not set.
            collect_artifacts: Callback to collect artifacts.
                Takes two named arguments:

                |  **project** (:class:`.VivadoProject`): The project that is being built.

                |  **output_path** (pathlib.Path): Where the build artifacts should be placed.

                | Must return True.
            kwargs: Other arguments as accepted by :meth:`.VivadoProject.build`.

                .. Note::
                    Argument ``project_path`` can not be set, it is set by this class
                    based on the ``project_paths`` argument to this function.

                    Argument ``num_threads`` is set by the ``num_threads_per_build``
                    argument to this function. This naming difference is done to avoid
                    confusion with regards to ``num_parallel_builds``.

        Return:
            True if everything went well.
        """
        if collect_artifacts:
            thread_safe_collect_artifacts = ThreadSafeCollectArtifacts(
                collect_artifacts
            ).collect_artifacts
        else:
            thread_safe_collect_artifacts = None

        build_wrappers = []
        for project in self.projects:
            project_output_path = self.get_build_project_output_path(
                project=project, projects_path=projects_path, output_path=output_path
            )

            build_wrapper = BuildProjectBuildWrapper(
                project=project,
                collect_artifacts=thread_safe_collect_artifacts,
                output_path=project_output_path,
                num_threads=num_threads_per_build,
                **kwargs,
            )
            build_wrappers.append(build_wrapper)

        return self._run_build_wrappers(
            projects_path=projects_path,
            build_wrappers=build_wrappers,
            num_parallel_builds=num_parallel_builds,
        )

    @staticmethod
    def get_build_project_output_path(
        project: "VivadoProject", projects_path: Path, output_path: Optional[Path] = None
    ) -> Path:
        """
        Find where build artifacts will be placed for a project.
        Arguments are the same as for :meth:`.build`.
        """
        if output_path:
            return output_path.resolve() / project.name

        return projects_path / project.name

    def open(self, projects_path: Path) -> bool:
        """
        Open the projects in EDA GUI.

        Arguments:
            projects_path: The projects are placed here.

        Return:
            True if everything went well.
        """
        build_wrappers = []
        for project in self.projects:
            build_wrappers.append(BuildProjectOpenWrapper(project))

        return self._run_build_wrappers(
            projects_path=projects_path,
            build_wrappers=build_wrappers,
            # For open there is no performance limitation. Set a high value.
            num_parallel_builds=20,
        )

    def _run_build_wrappers(
        self,
        projects_path: Path,
        build_wrappers: Union[
            list["BuildProjectCreateWrapper"],
            list["BuildProjectBuildWrapper"],
            list["BuildProjectOpenWrapper"],
        ],
        num_parallel_builds: int,
    ) -> bool:
        if not build_wrappers:
            # Return straight away if no builds are supplied
            return True

        start_time = time.time()

        color_printer = NO_COLOR_PRINTER if self._no_color else COLOR_PRINTER
        report = BuildReport(printer=color_printer)

        test_list = TestList()
        for build_wrapper in build_wrappers:
            test_list.add_test(build_wrapper)

        verbosity = BuildRunner.VERBOSITY_QUIET
        test_runner = BuildRunner(
            report=report,
            output_path=projects_path,
            verbosity=verbosity,
            num_threads=num_parallel_builds,
        )
        test_runner.run(test_list)

        all_builds_ok: bool = report.all_ok()
        report.set_real_total_time(time.time() - start_time)

        # True if the builds are for the "build" step (not "create" or "open")
        builds_are_build_step = isinstance(build_wrappers[0], BuildProjectBuildWrapper)
        # If we are building, we should print the summary that is at the end of the console output.
        # (however if we are creating or opening a project we should not print anything extra).
        # However if anything has failed, we should also print.
        if builds_are_build_step:
            # The length of the build summary depends on if we are working with netlist builds or
            # regular ones, so set the length given by one of the project objects.
            # Ignore typing error that 'Create' or 'Open' build wrappers do not have
            # 'build_result_report_length'. We only end up here if it is a Build.
            report_length_lines = build_wrappers[
                0
            ].build_result_report_length  # type: ignore[union-attr]
            report.set_report_length(report_length_lines=report_length_lines)

        # If all are OK then we should print the resource utilization numbers.
        # If not, then we print a few last lines of the log output.
        if builds_are_build_step or not all_builds_ok:
            report.print_str()

        return all_builds_ok

    def _iterate_projects(
        self, project_filters: list[str], include_netlist_not_top_builds: bool
    ) -> Iterable["VivadoProject"]:
        available_projects = []
        for module in self._modules:
            available_projects += module.get_build_projects()

        for project in available_projects:
            if project.is_netlist_build == include_netlist_not_top_builds:
                if not project_filters:
                    yield project

                else:
                    for project_filter in project_filters:
                        if fnmatch.filter([project.name], project_filter):
                            yield project

                            # Do not continue with further filters if we have already matched this
                            # project.
                            # Multiple filters might match the same project, and multiple objects
                            # of the same project will break build
                            break


class BuildProjectCreateWrapper:
    """
    Wrapper to create a build project, for usage in the build runner.
    Mimics a VUnit test object.
    """

    def __init__(self, project: "VivadoProject", **kwargs: Any) -> None:
        self.name = project.name
        self._project = project
        self._create_arguments = kwargs

    def run(self, output_path: Path, read_output: Any) -> bool:  # pylint: disable=unused-argument
        """
        VUnit test runner sends another argument "read_output" which we don't use.
        """
        this_projects_path = Path(output_path) / "project"
        return self._project.create(project_path=this_projects_path, **self._create_arguments)


class BuildProjectBuildWrapper:
    """
    Wrapper to build a project, for usage in the build runner.
    Mimics a VUnit test object.
    """

    def __init__(
        self,
        project: "VivadoProject",
        collect_artifacts: Optional[Callable[..., bool]],
        **kwargs: Any,
    ) -> None:
        self.name = project.name
        self._project = project
        self._collect_artifacts = collect_artifacts
        self._build_arguments = kwargs

    def run(self, output_path: Path, read_output: Any) -> bool:  # pylint: disable=unused-argument
        """
        VUnit test runner sends another argument "read_output" which we don't use.
        """
        this_projects_path = Path(output_path) / "project"
        build_result = self._project.build(project_path=this_projects_path, **self._build_arguments)

        if not build_result.success:
            self._print_build_result(build_result)
            return build_result.success

        # Proceed to artifact collection only if build succeeded.
        if self._collect_artifacts:
            if not self._collect_artifacts(
                project=self._project, output_path=self._build_arguments["output_path"]
            ):
                build_result.success = False

        # Print size at the absolute end
        self._print_build_result(build_result=build_result)
        return build_result.success

    @staticmethod
    def _print_build_result(build_result: "build_result.BuildResult") -> None:
        build_report = build_result.report()
        if build_report:
            # Add an empty line before the build result report, to have margin in how many lines are
            # printed. See the comments in BuildResult for an explanation.
            print()
            print(build_report)

    @property
    def build_result_report_length(self) -> int:
        """
        The number of lines in the build_result report from this project.
        """
        # The size summary, as returned by tsfpga.vivado.project.BuildResult is a JSON formatted
        # string with one line for each utilization category.
        # For Xilinx 7 series, there are 8 categories (Total LUTs, Logic LUTs, LUTRAMs,
        # SRLs, FFs, RAMB36, RAMB18, DSP Blocks). For UltraScale series there is one
        # extra (URAM).
        # Additionally, the size summary contains three extra lines for JSON braces and a title.
        #
        # This value is enough lines so the whole summary gets printed to console.
        # For 7 series, this will mean an extra blank line before the summary.
        #
        # This is a hack. Works for now, but is far from reliable.
        length_of_size_report = 3 + 8 + 1

        if self._project.is_netlist_build:
            # The logic level distribution report is five lines, plus a title line.
            # This report is only printed for netlist builds, where there is no configured clock
            # present. If there were many clocks present in the build, the report would be longer.
            length_of_logic_level_report = 5 + 1
            return length_of_size_report + length_of_logic_level_report

        return length_of_size_report


class BuildProjectOpenWrapper:
    """
    Wrapper to open a build project, for usage in the build runner.
    Mimics a VUnit test object.
    """

    def __init__(self, project: "VivadoProject") -> None:
        self.name = project.name
        self._project = project

    def run(self, output_path: Path, read_output: Any) -> bool:  # pylint: disable=unused-argument
        """
        VUnit test runner sends another argument "read_output" which we don't use.
        """
        this_projects_path = Path(output_path) / "project"
        return self._project.open(project_path=this_projects_path)


class BuildRunner(TestRunner):
    """
    Build runner that mimics a VUnit TestRunner. Most things are used as they are in the
    base class, but some behavior is overridden.
    """

    def _create_test_mapping_file(self, test_suites: Any) -> None:
        """
        Overloaded from super class.

        Do not create this file.

        We do not need it since folder name is the same as project name.
        """

    def _get_output_path(self, test_suite_name: str) -> str:
        """
        Overloaded from super class.

        Output folder name is the same as the project name.

        Original function adds a hash at the end of the folder name.
        We do not want that necessarily.
        """
        return str(Path(self._output_path) / test_suite_name)

    @staticmethod
    def _prepare_test_suite_output_path(output_path: str) -> None:
        """
        Overloaded from super class.

        Create the directory unless it already exists.

        Original function wipes the path before running a test. We do not want to do that
        since e.g. a Vivado project takes a long time to create and might contain a state
        that the user wants to keep.
        """
        create_directory(Path(output_path), empty=False)


class ThreadSafeCollectArtifacts:
    """
    A thread-safe wrapper around a user-supplied function that makes sure the function
    is not launched more than once at the same time. When two builds finish at the
    same time, race conditions can arise depending on what the function does.

    Note that this is a VERY fringe case, since builds usually take >20 minutes, and the
    collection probably only takes a few seconds. But it happens sometimes with the tsfpga
    example projects which are identical and quite fast (roughly three minutes).
    """

    def __init__(self, collect_artifacts: Callable[..., bool]) -> None:
        self._collect_artifacts = collect_artifacts
        self._lock = Lock()

    def collect_artifacts(self, project: "VivadoProject", output_path: Path) -> bool:
        with self._lock:
            return self._collect_artifacts(project=project, output_path=output_path)


class BuildReport(TestReport):
    def add_result(self, *args: Any, **kwargs: Any) -> None:
        """
        Overloaded from super class.

        Add a a test result.

        Uses a different Result class than the super method.
        """
        result = BuildResult(*args, **kwargs)
        self._test_results[result.name] = result
        self._test_names_in_order.append(result.name)

    def set_report_length(self, report_length_lines: int) -> None:
        """
        Set the report length for all test results that have been added to the report.
        """
        for test_result in self._test_results.values():
            test_result.set_report_length(report_length_lines)

    def print_latest_status(self, total_tests: int) -> None:
        """
        Overloaded from super class.

        This method is called for each build when it should print its result just as it finished,
        but other builds may not be finished yet.

        Inherited and adapted from the VUnit function:
        * Removed support for the "skipped" result.
        * Do not use abbreviations in the printout.
        * Use f-strings.
        """
        result = self._last_test_result()
        passed, failed, _ = self._split()

        if result.passed:
            self._printer.write("pass", fg="gi")
        elif result.failed:
            self._printer.write("fail", fg="ri")
        else:
            assert False

        count_summary = f"pass={len(passed)} fail={len(failed)} total={total_tests}"
        self._printer.write(f" ({count_summary}) {result.name} ({result.time:.1f} seconds)\n")


class BuildResult(TestResult):
    report_length_lines = None

    def _print_output(self, printer: Any, num_lines: int) -> None:
        """
        Print the last lines from the output file.

        The ``printer`` argument should of type ``ColorPrinter`` from VUnit.
        """
        output_tail = read_last_lines_of_file(Path(self._output_file_name), num_lines=num_lines)
        printer.write(output_tail)

    def set_report_length(self, report_length_lines: int) -> None:
        """
        Set how many lines shall be printed when this result is printed.
        """
        self.report_length_lines = report_length_lines

    def print_status(self, printer: Any, padding: int = 0) -> None:
        """
        Overloaded from super class.

        The ``printer`` argument should of type ``ColorPrinter`` from VUnit.

        This method is called for each build when it should print its result in the "Summary" at
        the end when all builds have finished.

        Inherited and adapted from the VUnit function.
        """
        if self.passed and self.report_length_lines is not None:
            # Build passed, print build summary of the specified length. The length is only
            # set if this is a "build" result (not "create" or "open").
            self._print_output(printer=printer, num_lines=self.report_length_lines)
        else:
            # The build failed, which can either be caused by
            # 1. IDE build failure
            # 2. IDE build succeeded, but post build hook, or size checkers failed.
            # 3. Other python error (directory already exists, ...)
            # In the case of IDE build failed, we want a significant portion of the output, to be
            # able to see an indication of what failed. In the case of size checkers, we want to see
            # all the printouts from all checkers, to see which one failed. Since there are at most
            # eight resource categories, it is reasonable to assume that there will never be more
            # than eight size checkers.
            self._print_output(printer=printer, num_lines=25)

        # Print the regular output from the VUnit class.
        # A little extra margin between build name and execution time makes the output more readable
        super().print_status(printer=printer, padding=padding + 2)
        # Add an empty line between each build, for readability.
        printer.write("\n")



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------


# Standard libraries
from typing import Union


class StringGenericValue:
    """
    Use this type for generic values of type ``string``.
    """

    def __init__(self, value: str) -> None:
        """
        Arguments:
            value: A string of variable length with any content.
        """
        if not isinstance(value, str):
            raise ValueError(
                f"Expected {self.__class__.__name__} value to be of type str."
                f' Got type="{type(value)}", value="{value}".'
            )

        if " " in value:
            raise ValueError(
                f'Expected {self.__class__.__name__} value to not contain spaces. Got "{value}".'
            )

        self.value = value

    def __str__(self) -> str:
        return self.value


class BitVectorGenericValue:
    """
    Use this type for generic values of type ``std_logic_vector``.
    """

    def __init__(self, value: str) -> None:
        """
        Arguments:
            value: A string of variable length containing only "1" or "0".
        """
        if not isinstance(value, str):
            raise ValueError(
                f"Expected {self.__class__.__name__} value to be of type str."
                f' Got type="{type(value)}", value="{value}".'
            )

        for bit_value in value:
            if bit_value not in ["1", "0"]:
                raise ValueError(
                    f'Expected {self.__class__.__name__} value to contain only "1" or "0".'
                    f' Got "{value}".'
                )

        self.value = value

    @property
    def length(self) -> int:
        """
        The number of bits in the vector.
        """
        return len(self.value)

    def __str__(self) -> str:
        return self.value


def get_vivado_tcl_generic_value(
    value: Union[bool, int, float, StringGenericValue, BitVectorGenericValue]
) -> str:
    """
    Convert generic values of different types to the format recognized by Vivado TCL:
    https://www.xilinx.com/support/answers/52217.html

    Arguments:
        value: A generic value of native Python type..

    Return:
        The ``value`` formatted as TCL.
    """
    # Note that bool is a sub-class of int in Python, so check for bool must be first
    if isinstance(value, bool):
        return f"1'b{int(value)}"

    if isinstance(value, int):
        return str(value)

    if isinstance(value, float):
        return str(value)

    if isinstance(value, BitVectorGenericValue):
        return f"{value.length}'b{value.value}"

    if isinstance(value, StringGenericValue):
        return f'"{value.value}"'

    message = f'Unsupported type for generic. Got type="{type(value)}", value="{value}".'

    # When the type is a string, we can be a little more helpful and indicate what types shall
    # be used instead.
    if isinstance(value, str):
        message += (
            " Please use either of the explicit types StringGenericValue or BitVectorGenericValue."
        )

    raise ValueError(message)



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from pathlib import Path
from unittest.mock import MagicMock, patch

# Third party libraries
import pytest

# First party libraries
from tsfpga.test.test_utils import file_contains_string
from tsfpga.vivado.simlib import VivadoSimlib


# pylint: disable=redefined-outer-name
@pytest.fixture
def simlib_test(tmp_path):
    class SimlibCommercialTestFixture:
        def __init__(self):
            self.output_path = tmp_path / "simlib"

            self.simulator_prefix = "/opt/Aldec/Riviera-PRO-2018.10-x64/bin"
            self.vivado_path = Path("/tools/xilinx/Vivado/2019.2/bin/vivado")

            self.vivado_simlib = self.get_vivado_simlib(self.simulator_prefix, self.vivado_path)

        def get_vivado_simlib(
            self, simulator_prefix, vivado_path, simulator_class_name="rivierapro"
        ):
            simulator_class = MagicMock()
            simulator_class.name = simulator_class_name
            simulator_class.find_prefix.return_value = simulator_prefix

            vunit_proj = MagicMock()
            vunit_proj._simulator_class = simulator_class  # pylint: disable=protected-access

            return VivadoSimlib.init(self.output_path, vunit_proj, vivado_path)

        @staticmethod
        def assert_should_compile(vivado_simlib):
            assert vivado_simlib.compile_is_needed
            with patch("tsfpga.vivado.simlib_commercial.run_vivado_tcl", autospec=True) as mock:
                vivado_simlib.compile_if_needed()
                mock.assert_called_once()

        @staticmethod
        def assert_should_not_compile(vivado_simlib):
            assert not vivado_simlib.compile_is_needed
            with patch("tsfpga.vivado.simlib_commercial.run_vivado_tcl", autospec=True) as mock:
                vivado_simlib.compile_if_needed()
                mock.assert_not_called()

    return SimlibCommercialTestFixture()


def test_should_not_recompile(simlib_test):
    simlib_test.assert_should_compile(simlib_test.vivado_simlib)
    simlib_test.assert_should_not_compile(simlib_test.vivado_simlib)


def test_new_simulator_version_should_cause_recompile(simlib_test):
    simlib_test.assert_should_compile(simlib_test.vivado_simlib)
    simlib_test.assert_should_not_compile(simlib_test.vivado_simlib)

    vivado_simlib = simlib_test.get_vivado_simlib(
        simulator_prefix="/opt/Aldec/Riviera-PRO-1975.01-x64/bin",
        vivado_path=simlib_test.vivado_path,
    )
    simlib_test.assert_should_compile(vivado_simlib)
    simlib_test.assert_should_not_compile(vivado_simlib)


def test_new_vivado_version_should_cause_recompile(simlib_test):
    simlib_test.assert_should_compile(simlib_test.vivado_simlib)
    simlib_test.assert_should_not_compile(simlib_test.vivado_simlib)

    vivado_simlib = simlib_test.get_vivado_simlib(
        simulator_prefix=simlib_test.simulator_prefix,
        vivado_path=Path("/tools/xilinx/Vivado/1337.2/bin/vivado"),
    )
    simlib_test.assert_should_compile(vivado_simlib)
    simlib_test.assert_should_not_compile(vivado_simlib)


def test_remapping_of_vivado_simulator_names(simlib_test):
    def check_simulator_name(vivado_simlib, name):
        with patch("tsfpga.vivado.simlib_commercial.run_vivado_tcl", autospec=True) as _:
            vivado_simlib.compile_if_needed()

        assert file_contains_string(
            file=vivado_simlib.output_path / "compile_simlib.tcl",
            string=f"compile_simlib -simulator {name} -simulator_exec_path ",
        )

    # Regular Siemens ModelSim should be called "modelsim"
    vivado_simlib = simlib_test.get_vivado_simlib(
        simulator_prefix=Path("/opt/intelFPGA/20.1/modelsim_ase/bin/"),
        vivado_path=Path("/tools/xilinx/Vivado/2019.2/bin/vivado"),
        simulator_class_name="modelsim",
    )
    check_simulator_name(vivado_simlib=vivado_simlib, name="modelsim")

    # Siemens Questa is called "modelsim" in VUnit but Vivado needs the name "questasim".
    vivado_simlib = simlib_test.get_vivado_simlib(
        simulator_prefix=Path("/opt/intelFPGA_pro/22.2/questa_fse/bin"),
        vivado_path=Path("/tools/xilinx/Vivado/2019.2/bin/vivado"),
        simulator_class_name="modelsim",
    )
    check_simulator_name(vivado_simlib=vivado_simlib, name="questasim")

    # Aldec Riviera-PRO is called "rivierapro" in VUnit but Vivado needs the name "riviera"
    vivado_simlib = simlib_test.get_vivado_simlib(
        simulator_prefix=Path("/opt/Aldec/Riviera-PRO-1975.01-x64/bin"),
        vivado_path=Path("/tools/xilinx/Vivado/2019.2/bin/vivado"),
        simulator_class_name="rivierapro",
    )
    check_simulator_name(vivado_simlib=vivado_simlib, name="riviera")



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import io
from unittest.mock import Mock, patch

# Third party libraries
import pytest

# First party libraries
from tsfpga.vivado.build_result import BuildResult
from tsfpga.vivado.build_result_checker import (
    DspBlocks,
    EqualTo,
    LessThan,
    MaximumLogicLevel,
    Ramb,
    Ramb36,
    TotalLuts,
)


@patch("sys.stdout", new_callable=io.StringIO)
def test_size_checker_less_than_pass(mock_stdout):
    build_result = BuildResult(name="")
    build_result.synthesis_size = {"Total LUTs": 5}

    assert TotalLuts(LessThan(10)).check(build_result)
    assert mock_stdout.getvalue() == "Result check passed for Total LUTs: 5 (< 10)\n"


@patch("sys.stdout", new_callable=io.StringIO)
def test_size_checker_less_than_fail(mock_stdout):
    build_result = BuildResult(name="")
    build_result.synthesis_size = {"Total LUTs": 5}

    assert not TotalLuts(LessThan(5)).check(build_result)
    assert mock_stdout.getvalue() == "Result check failed for Total LUTs. Got 5, expected < 5.\n"


@patch("sys.stdout", new_callable=io.StringIO)
def test_size_checker_equal_to_pass(mock_stdout):
    build_result = BuildResult(name="")
    build_result.synthesis_size = {"RAMB36": 0, "Total LUTs": 5}

    assert Ramb36(EqualTo(0)).check(build_result)
    assert TotalLuts(EqualTo(5)).check(build_result)

    assert mock_stdout.getvalue() == (
        "Result check passed for RAMB36: 0 (0)\nResult check passed for Total LUTs: 5 (5)\n"
    )


@patch("sys.stdout", new_callable=io.StringIO)
def test_size_checker_equal_to_fail(mock_stdout):
    build_result = BuildResult(name="")
    build_result.synthesis_size = {"Total LUTs": 5}

    assert not TotalLuts(EqualTo(50)).check(build_result)
    assert mock_stdout.getvalue() == "Result check failed for Total LUTs. Got 5, expected 50.\n"


def test_size_checker_dsp_blocks_has_two_names():
    build_result = BuildResult(name="")

    build_result.synthesis_size = {"DSP Blocks": 5}
    assert DspBlocks(EqualTo(5)).check(build_result)
    assert not DspBlocks(EqualTo(2)).check(build_result)

    build_result.synthesis_size = {"DSP48 Blocks": 7}
    assert DspBlocks(EqualTo(7)).check(build_result)
    assert not DspBlocks(EqualTo(2)).check(build_result)


def test_size_checker_aggregated_ramb():
    build_result = BuildResult(name="")

    build_result.synthesis_size = {"RAMB18": 5, "RAMB36": 20}
    assert Ramb(EqualTo(22.5)).check(build_result)
    assert not Ramb(EqualTo(22)).check(build_result)

    build_result.synthesis_size = {"RAMB18": 5, "RAMB36": 0}
    assert Ramb(EqualTo(2.5)).check(build_result)

    build_result.synthesis_size = {"RAMB18": 0, "RAMB36": 20}
    assert Ramb(EqualTo(20)).check(build_result)


def test_size_checker_raises_exception_if_synthesis_size_is_not_set():
    build_result = BuildResult(name="")
    build_result.synthesis_size = None

    with pytest.raises(ValueError) as exception_info:
        assert Ramb36(EqualTo(0)).check(build_result)
    assert str(exception_info.value) == "Should only call after successful synthesis"


def test_size_checker_raises_exception_if_synthesis_size_does_not_have_requested_resource():
    build_result = BuildResult(name="")
    build_result.synthesis_size = {"Total LUTs": 5}

    with pytest.raises(ValueError) as exception_info:
        assert Ramb36(EqualTo(0)).check(build_result)
    assert (
        str(exception_info.value)
        == 'Synthesis result size does not contain the requested resource: "RAMB36"'
    )


@patch("sys.stdout", new_callable=io.StringIO)
def test_logic_level_checker_equal_to_pass(mock_stdout):
    build_result = Mock(spec=BuildResult)
    build_result.maximum_logic_level = 5

    assert MaximumLogicLevel(EqualTo(5)).check(build_result)
    assert mock_stdout.getvalue() == "Result check passed for maximum logic level: 5 (5)\n"


@patch("sys.stdout", new_callable=io.StringIO)
def test_logic_level_checker_less_than_fail(mock_stdout):
    build_result = Mock(spec=BuildResult)
    build_result.maximum_logic_level = 11
    assert not MaximumLogicLevel(LessThan(7)).check(build_result)
    assert (
        mock_stdout.getvalue()
        == "Result check failed for maximum logic level. Got 11, expected < 7.\n"
    )


def test_logic_level_checker_raises_exception_if_logic_level_is_not_set():
    build_result = BuildResult(name="")

    with pytest.raises(ValueError) as exception_info:
        MaximumLogicLevel(LessThan(7)).check(build_result)
    assert str(exception_info.value) == "Should only call after successful synthesis"



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------
# A set of methods for building sphinx docs. Should be reusable between projects.
# --------------------------------------------------------------------------------------------------

# Standard libraries
import sys
from datetime import datetime
from pathlib import Path
from subprocess import check_call
from typing import Iterable, Optional

# Third party libraries
from git.repo import Repo
from packaging.version import Version, parse

# First party libraries
from tsfpga.system_utils import read_file


def generate_release_notes(
    repo_root: Path, release_notes_directory: Path, project_name: str
) -> str:
    """
    Generate release notes in RST format based on a directory full of release note files.
    Will match each file to a git tag.

    Arguments:
        repo_root: Git commands will be executed here.
        release_notes_directory: Location of release notes files.
        project_name: Name of project will be used for the GitHub link.

    Return:
        RST code with release notes.
    """
    rst = ""

    for release, previous_release_git_tag in _get_release_notes_files(
        repo_root=repo_root, release_notes_directory=release_notes_directory
    ):
        heading = f"{release.version} ({release.date})"
        rst += heading + "\n"
        rst += "-" * len(heading) + "\n"
        rst += "\n"
        if previous_release_git_tag is not None:
            diff_url = (
                f"https://github.com/{project_name}/{project_name}/compare/"
                f"{previous_release_git_tag}...{release.git_tag}"
            )
            rst += f"`Changes since previous release <{diff_url}>`__\n"
        rst += "\n"
        rst += read_file(release.release_notes_file)
        rst += "\n"

    return rst


def _get_release_notes_files(
    repo_root: Path, release_notes_directory: Path
) -> Iterable[tuple["Release", Optional[str]]]:
    """
    Iterate the release notes.
    """
    unreleased_notes_file = release_notes_directory / "unreleased.rst"

    release_notes = []

    # Get all versioned release notes files and sort them in order newest -> oldest
    for release_notes_file in release_notes_directory.glob("*.rst"):
        if not release_notes_file == unreleased_notes_file:
            release_notes.append(release_notes_file)

    # Sort by parsing the version number in the file name. Newest to oldest.
    def sort_key(path: Path) -> Version:
        return parse(path.stem)

    release_notes.sort(key=sort_key, reverse=True)

    # The "Unreleased" shall be first
    release_notes.insert(0, unreleased_notes_file)

    repo = Repo(repo_root)
    releases = [
        Release(repo=repo, release_notes_file=release_notes_file)
        for release_notes_file in release_notes
    ]

    for idx, release in enumerate(releases):
        if idx == len(releases) - 1:
            previous_release_git_tag = None
        else:
            previous_release_git_tag = releases[idx + 1].git_tag

        yield release, previous_release_git_tag


class Release:
    """
    Used to represent a release.
    """

    def __init__(self, repo: Repo, release_notes_file: Path) -> None:
        self.release_notes_file = release_notes_file

        version = release_notes_file.stem
        if version == "unreleased":
            self.version = "Unreleased"
            self.git_tag = "main"
            self.date = "YYYY-MM-DD"
        else:
            self.version = version
            self.git_tag = "v" + self.version
            self.date = self.get_git_date_from_tag(repo=repo, tag=self.git_tag)

    @staticmethod
    def get_git_date_from_tag(repo: Repo, tag: str) -> str:
        """
        Get a formatted date string, gathered from git log based on tag name.
        """
        timestamp = repo.tag(f"refs/tags/{tag}").commit.committed_date
        time = datetime.fromtimestamp(timestamp)
        return f"{time.day} {time:%B} {time.year}".lower()


def build_sphinx(build_path: Path, output_path: Path) -> None:
    """
    Execute sphinx on command line to build HTML documentation.

    Arguments:
        build_path: The location that contains ``conf.py`` and ``index.rst``.
        output_path: Where to place the generated HTML.
    """
    # Since we set the working directory when making the system call, the paths must be absolute
    build_path = build_path.resolve()
    output_path = output_path.resolve()

    cmd = [
        sys.executable,
        "-m",
        "sphinx",
        # Turn warnings into errors.
        "-W",
        # Show full traceback upon error.
        "-T",
        str(build_path),
        str(output_path),
    ]
    check_call(cmd, cwd=build_path)

    index_html = output_path / "index.html"
    assert index_html.exists(), index_html
    print(f"Open with:\nfirefox {index_html} &")



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# First party libraries
from tsfpga.system_utils import create_file
from tsfpga.vhdl_file_documentation import VHDL_COMMENT_SEPARATOR, VhdlFileDocumentation


def test_documentation_header(tmp_path):
    data = f"""\
{VHDL_COMMENT_SEPARATOR}
-- Copyright bla bla
{VHDL_COMMENT_SEPARATOR}
-- This is my documentation
--
--     with
--     indentation
--
-- And empty lines.
{VHDL_COMMENT_SEPARATOR}

"""
    expected = """\
This is my documentation

    with
    indentation

And empty lines.
"""

    vhd_file = create_file(tmp_path / "file_for_test.vhd", data)
    assert VhdlFileDocumentation(vhd_file).get_header_rst() == expected


def test_only_copyright_header_should_return_no_documentation_header(tmp_path):
    data = f"""\
{VHDL_COMMENT_SEPARATOR}
-- Copyright bla bla
{VHDL_COMMENT_SEPARATOR}

"""

    vhd_file = create_file(tmp_path / "file_for_test.vhd", data)
    assert VhdlFileDocumentation(vhd_file).get_header_rst() is None


def run_get_symbolator_component_test(tmp_path, vhdl_code, expected):
    vhd_file_path = create_file(tmp_path / "test_entity.vhd", vhdl_code)
    got = VhdlFileDocumentation(vhd_file_path).get_symbolator_component()

    assert got == expected


def test_get_symbolator_component_simple(tmp_path):
    data = """
entity test_entity is
  generic (
    num_interfaces : positive
  );
  port (
    clk : in std_ulogic;
    --# {{}}
    input_ready : out std_ulogic := '0';
    input_valid : in std_ulogic;
    --# {{}}
    output_ready : in std_ulogic_vector(0 to num_interfaces - 1);
    output_valid : out std_ulogic_vector(0 to num_interfaces - 1) := (others => '0')
  );
end entity;
"""

    # entity->component, no ranges, and no default values.
    expected = """\
component test_entity is
  generic (
    num_interfaces : positive
  );
  port (
    clk : in std_ulogic;
    --# {{}}
    input_ready : out std_ulogic;
    input_valid : in std_ulogic;
    --# {{}}
    output_ready : in std_ulogic_vector;
    output_valid : out std_ulogic_vector
  );
end component;"""

    run_get_symbolator_component_test(tmp_path=tmp_path, vhdl_code=data, expected=expected)


def test_get_symbolator_component_complex(tmp_path):
    data = """
library common;
use common.addr_pkg.all;


entity test_entity is
  generic(
    buffer_segment_length_bytes : positive;
    dummy : positive := 10;
    -- comment
    silly : positive_vec_t(0 to 1) :=
      -- comment again
      (others => 4)
  )
  ;port(
    clk : in std_ulogic;
    --# {{}}
    request_ready : in std_ulogic := '0';
    request_valid : out std_ulogic_vector(my_range_constant) := '0';
    request_address : out addr_t(apa to hest) := (others => '0');
    request_funky : out addr_t(apa to hest) :=
      (others => '0');
    request_slinky : out addr_t(apa to hest) := (
      -- comment
      others => '0'
    );
    --# {{}}
    release_last_address : in std_ulogic_vector(enables'range)
      := '0';
    --# {{}}
    buffer_start_address : in addr_t(hest downto 0);
    buffer_last_address : in addr_t
      := (others => '0')
   );
end entity;

architecture a of test_entity is

begin

end architecture;
end;
end a;
"""

    # entity->component, no ranges, and no default values.
    expected = """\
component test_entity is
  generic (
    buffer_segment_length_bytes : positive;
    dummy : positive;
    silly : positive_vec_t
  );
  port (
    clk : in std_ulogic;
    --# {{}}
    request_ready : in std_ulogic;
    request_valid : out std_ulogic_vector;
    request_address : out addr_t;
    request_funky : out addr_t;
    request_slinky : out addr_t;
    --# {{}}
    release_last_address : in std_ulogic_vector;
    --# {{}}
    buffer_start_address : in addr_t;
    buffer_last_address : in addr_t
  );
end component;"""

    run_get_symbolator_component_test(tmp_path=tmp_path, vhdl_code=data, expected=expected)


def test_get_symbolator_component_with_attributes(tmp_path):
    data = """
entity test_entity is
  generic (
    length_attribute : positive;
    attribute_length : positive
  );
  port (
    data_attribute : out std_ulogic_vector(my_range_constant) := '0';
    attribute_data : out std_ulogic_vector(my_range_constant) := '0'
  );

  attribute apa : string;
  attribute apa of test_entity : entity is "hest";
end entity;
"""

    expected = """\
component test_entity is
  generic (
    length_attribute : positive;
    attribute_length : positive
  );
  port (
    data_attribute : out std_ulogic_vector;
    attribute_data : out std_ulogic_vector
  );
end component;"""

    run_get_symbolator_component_test(tmp_path=tmp_path, vhdl_code=data, expected=expected)


def test_get_symbolator_component_no_generics(tmp_path):
    data = """
entity test_entity is
  port (
    dummy_signal : out std_ulogic_vector(my_range_constant) := '0'
  );
end entity;
"""

    # entity->component, no ranges, and no default values.
    expected = """\
component test_entity is
  port (
    dummy_signal : out std_ulogic_vector
  );
end component;"""

    run_get_symbolator_component_test(tmp_path=tmp_path, vhdl_code=data, expected=expected)


def test_get_symbolator_component_last_port_no_newline(tmp_path):
    data = """
entity test_entity is
  generic (
    buffer_segment_length_bytes : positive_vec_t(0 to 1));
  port (
    dummy_signal : out std_ulogic_vector(my_range_constant) := '0');end entity;
"""

    # entity->component, no ranges, and no default values.
    expected = """\
component test_entity is
  generic (
    buffer_segment_length_bytes : positive_vec_t
  );
  port (
    dummy_signal : out std_ulogic_vector
  );
end component;"""

    run_get_symbolator_component_test(tmp_path=tmp_path, vhdl_code=data, expected=expected)


def test_get_symbolator_component_last_port_parenthesis_on_same_line(tmp_path):
    data = """
entity test_entity is
  generic (
    buffer_segment_length_bytes : positive_vec_t(0 to 1) := (1, 2)  )
  ;
  port (
    dummy_signal : out std_ulogic_vector(my_range_constant) := '0'  )
  ;
end entity;
"""

    # entity->component, no ranges, and no default values.
    expected = """\
component test_entity is
  generic (
    buffer_segment_length_bytes : positive_vec_t
  );
  port (
    dummy_signal : out std_ulogic_vector
  );
end component;"""

    run_get_symbolator_component_test(tmp_path=tmp_path, vhdl_code=data, expected=expected)


def test_get_symbolator_component_with_comments(tmp_path):
    data = """
entity test_entity is -- hello
  generic ( -- test ); end entity; -- Comment --in a --       comment
    apa : natural := 1-- ;
-- hest : natural
  );
  port (
    dummy : std_ulogic -- ); end entity;
    -- := 0
  );
end entity;
"""

    expected = """\
component test_entity is
  generic (
    apa : natural
  );
  port (
    dummy : std_ulogic
  );
end component;"""

    run_get_symbolator_component_test(tmp_path=tmp_path, vhdl_code=data, expected=expected)


def test_get_symbolator_component_with_separator_comments(tmp_path):
    data = """
entity test_entity is
  port (
    apa : out std_ulogic; -- trailing comment that shall be removed
    --# {{}}
    hest : out std_ulogic;
-- Out of place comment that shall be removed
-- foo : in std_ulogic;
    --# {{test 123}}
    zebra : out std_ulogic;
  );
end entity;
"""

    # entity->component, no ranges, and no default values.
    expected = """\
component test_entity is
  port (
    apa : out std_ulogic;
    --# {{}}
    hest : out std_ulogic;
    --# {{test 123}}
    zebra : out std_ulogic;
  );
end component;"""

    run_get_symbolator_component_test(tmp_path=tmp_path, vhdl_code=data, expected=expected)


def test_get_symbolator_component_with_complex_array_width(tmp_path):
    data = """
entity test_entity is
  port (
    dummy : out std_ulogic_vector(
      my_function(
        value=>generic_value
      ) - 1
      downto 0
    ) := (others => '0')
  );
end entity;
"""

    expected = """\
component test_entity is
  port (
    dummy : out std_ulogic_vector
  );
end component;"""

    run_get_symbolator_component_test(tmp_path=tmp_path, vhdl_code=data, expected=expected)


def run_get_symbolator_component_end_test(tmp_path, end_statement):
    data = f"""
entity test_entity is
  port (
    clk : in std_ulogic;
    dummy_signal : out std_ulogic_vector(my_range_constant) := '0'
   );
{end_statement};

architecture a of test_entity is
begin

end architecture;
end;
end a;
"""

    # entity->component, no ranges, and no default values.
    expected = """\
component test_entity is
  port (
    clk : in std_ulogic;
    dummy_signal : out std_ulogic_vector
  );
end component;"""

    run_get_symbolator_component_test(tmp_path=tmp_path, vhdl_code=data, expected=expected)


def run_get_symbolator_component_end_test_with_whitespace(tmp_path, end_statement):
    # As above but with lots of whitespace around the keywords.
    data = f"""
entity
 test_entity
  is

  port (
    clk : in std_ulogic;
    dummy_signal : out std_ulogic_vector(my_range_constant) := '0'
  )
   ;


 {end_statement}
  ;

architecture a of test_entity is
begin

end architecture;
end;
end a;
"""

    # entity->component, no ranges, and no default values.
    expected = """\
component test_entity is
  port (
    clk : in std_ulogic;
    dummy_signal : out std_ulogic_vector
  );
end component;"""

    run_get_symbolator_component_test(tmp_path=tmp_path, vhdl_code=data, expected=expected)


def test_get_symbolator_component_end_only_keyword(tmp_path):
    end_statement = "end"
    run_get_symbolator_component_end_test(tmp_path=tmp_path, end_statement=end_statement)


def test_get_symbolator_component_no_end_entity_with_whitespace(tmp_path):
    end_statement = "end   \n"
    run_get_symbolator_component_end_test_with_whitespace(
        tmp_path=tmp_path, end_statement=end_statement
    )


def test_get_symbolator_component_end_name(tmp_path):
    end_statement = "end test_entity"
    run_get_symbolator_component_end_test(tmp_path=tmp_path, end_statement=end_statement)


def test_get_symbolator_component_end_name_with_whitespace(tmp_path):
    end_statement = "end \n  test_entity  \n"
    run_get_symbolator_component_end_test_with_whitespace(
        tmp_path=tmp_path, end_statement=end_statement
    )


def test_get_symbolator_component_end_entity_name(tmp_path):
    end_statement = "end entity test_entity"
    run_get_symbolator_component_end_test(tmp_path=tmp_path, end_statement=end_statement)


def test_get_symbolator_component_end_entity_name_with_whitespace(tmp_path):
    end_statement = "end \n  entity  \n test_entity  \n"
    run_get_symbolator_component_end_test_with_whitespace(
        tmp_path=tmp_path, end_statement=end_statement
    )



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from pathlib import Path
from shutil import which
from typing import Optional

# Third party libraries
from vunit.ostools import Process

# First party libraries
from tsfpga.git_utils import get_git_sha
from tsfpga.math_utils import to_binary_string


def run_vivado_tcl(vivado_path: Optional[Path], tcl_file: Path, no_log_file: bool = False) -> bool:
    """
    Setting cwd ensures that any .log or .jou files produced are placed in
    the same directory as the TCL file that produced them.

    Arguments:
        vivado_path: Path to Vivado executable. Can set to ``None``
            to use whatever version is in ``PATH``.
        tcl_file: Path to TCL file.
        no_log_file: Optionally set Vivado flags to not create log and journal files.

    Return:
        True if everything went well.
    """
    tcl_file = tcl_file.resolve()

    cmd = [
        str(get_vivado_path(vivado_path)),
        "-mode",
        "batch",
        "-notrace",
        "-source",
        str(tcl_file),
    ]
    if no_log_file:
        cmd += ["-nojournal", "-nolog"]

    try:
        Process(args=cmd, cwd=tcl_file.parent).consume_output()
    except Process.NonZeroExitCode:
        return False
    return True


def run_vivado_gui(vivado_path: Optional[Path], project_file: Path) -> bool:
    """
    Setting cwd ensures that any .log or .jou files produced are placed in
    the same directory as the project.

    Arguments:
        vivado_path: Path to Vivado executable. Can set to ``None``
            to use whatever version is in ``PATH``.
        project_file: Path to a project .xpr file.

    Return:
        True if everything went well.
    """
    project_file = project_file.resolve()
    if not project_file.exists():
        raise FileNotFoundError(f"Project does not exist: {project_file}")

    cmd = [str(get_vivado_path(vivado_path)), "-mode", "gui", str(project_file)]

    try:
        Process(args=cmd, cwd=project_file.parent).consume_output()
    except Process.NonZeroExitCode:
        return False
    return True


def get_vivado_path(vivado_path: Optional[Path] = None) -> Path:
    """
    Wrapper to get a path to Vivado executable.

    Arguments:
        vivado_path: Path to vivado executable. Set to ``None`` to use whatever
            is available in the system ``PATH``.
    """
    if vivado_path is not None:
        return vivado_path.resolve()

    which_vivado = which("vivado")
    if which_vivado is None:
        raise FileNotFoundError("Could not find vivado on PATH")

    return Path(which_vivado).resolve()


def get_vivado_version(vivado_path: Optional[Path] = None) -> str:
    """
    Get the version number of the Vivado installation.

    Arguments:
        vivado_path: Path to vivado executable. Set to ``None`` to use whatever
            is available in the system ``PATH``.

    Return:
        The version, e.g. ``"2021.2"``.
    """
    vivado_path = get_vivado_path(vivado_path=vivado_path)

    # E.g. "/home/lukas/work/Xilinx/Vivado/2021.2/bin/vivado" -> "2021.2"
    vivado_version = vivado_path.parent.parent.name

    return vivado_version


def get_git_sha_slv(git_directory: Path) -> tuple[str, str]:
    """
    Get the current git SHA encoded as 32-character binary strings. Suitable for setting
    as ``std_logic_vector`` generics to a Vivado build, where they can be assigned to
    software-accessible registers to keep track of what revision an FPGA was built from.

    Will return the left-most 16 characters of the git SHA, encoded into two 32-character strings.

    Arguments:
        directory (pathlib.Path): The directory where git commands will be run.

    Return:
        First object in tuple is the left-most eight characters of the git SHA
        encoded as a 32-character binary string. Second object is the next eight characters from
        the git SHA.
    """
    git_sha = get_git_sha(directory=git_directory)
    assert len(git_sha) == 16

    def hex_to_binary_string(hex_string: str) -> str:
        assert len(hex_string) == 8
        int_value = int(hex_string, base=16)

        return to_binary_string(value=int_value, result_width=32)

    left_binary_string = hex_to_binary_string(hex_string=git_sha[0:8])
    right_binary_string = hex_to_binary_string(hex_string=git_sha[8:16])

    return (left_binary_string, right_binary_string)


def to_tcl_path(path: Path) -> str:
    """
    Return a path string in a format suitable for TCL.
    """
    return str(path.resolve()).replace("\\", "/")



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import argparse
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Optional

# Third party libraries
from hdl_registers.generator.c.header import CHeaderGenerator
from hdl_registers.generator.cpp.header import CppHeaderGenerator
from hdl_registers.generator.cpp.implementation import CppImplementationGenerator
from hdl_registers.generator.cpp.interface import CppInterfaceGenerator
from hdl_registers.generator.html.page import HtmlPageGenerator
from hdl_registers.generator.python.pickle import PythonPickleGenerator

# First party libraries
from tsfpga.build_project_list import BuildProjectList
from tsfpga.system_utils import create_directory

if TYPE_CHECKING:
    # First party libraries
    from tsfpga.module_list import ModuleList
    from tsfpga.vivado.project import VivadoProject


def arguments(default_temp_dir: Path) -> argparse.Namespace:
    """
    Setup of arguments for the example build flow.

    Arguments:
        default_temp_dir (pathlib.Path): Default value for output paths.
    """
    parser = argparse.ArgumentParser(
        "Create, synth and build an FPGA project",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    group = parser.add_mutually_exclusive_group()

    group.add_argument("--list-only", action="store_true", help="list the available projects")

    group.add_argument(
        "--generate-registers-only",
        action="store_true",
        help="only generate the register artifacts (C/C++ code, HTML, ...) for inspection",
    )

    group.add_argument("--create-only", action="store_true", help="only create projects")

    group.add_argument("--synth-only", action="store_true", help="only synthesize projects")

    group.add_argument(
        "--from-impl",
        action="store_true",
        help="run impl and onwards on an existing synthesized projects",
    )

    group.add_argument("--open", action="store_true", help="open existing projects in the GUI")

    group.add_argument(
        "--collect-artifacts-only",
        action="store_true",
        help="collect artifacts of previously successful builds",
    )

    parser.add_argument(
        "--use-existing-project",
        action="store_true",
        help="build existing projects, or create first if they do not exist",
    )

    parser.add_argument(
        "--netlist-builds",
        action="store_true",
        help="use netlist build projects instead of top level build projects",
    )

    parser.add_argument(
        "--projects-path",
        type=Path,
        default=default_temp_dir / "projects",
        help="the FPGA build projects will be placed here",
    )

    parser.add_argument(
        "--ip-cache-path",
        type=Path,
        default=default_temp_dir / "vivado_ip_cache",
        help="location of Vivado IP cache",
    )

    parser.add_argument(
        "--output-path",
        type=Path,
        required=False,
        help="the output products (bit file, ...) will be placed here",
    )

    parser.add_argument(
        "--num-parallel-builds", type=int, default=8, help="Number of parallel builds to launch"
    )

    parser.add_argument(
        "--num-threads-per-build",
        type=int,
        default=4,
        help="number of threads for each build process",
    )

    parser.add_argument("--no-color", action="store_true", help="disable color in printouts")

    parser.add_argument(
        "project_filters",
        nargs="*",
        help="filter for which projects to build. Can use wildcards. Leave empty for all.",
    )

    args = parser.parse_args()

    assert (
        args.use_existing_project or not args.from_impl
    ), "Must set --use-existing-project when using --from-impl"

    return args


def setup_and_run(  # pylint: disable=too-many-return-statements
    modules: "ModuleList",
    projects: BuildProjectList,
    args: argparse.Namespace,
    collect_artifacts_function: Optional[Callable[["VivadoProject", Path], bool]],
) -> int:
    """
    Setup and execute build projects.
    As instructed by the arguments.

    Arguments:
        modules: When running a register generation, registers from these
            modules will be included.
        projects: These build projects will be built.
        args: Command line argument namespace.
        collect_artifacts_function: Function pointer to a function that collects build artifacts.
            Will be run after a successful implementation build.
            The function must return ``True`` if successful and ``False`` otherwise.
            It will receive the ``project`` and ``output_path`` as arguments.
            Can be ``None`` if no special artifact collection operation shall be run.

    Return:
        0 if everything passed, otherwise non-zero.
        Can be used for system exit code.
    """
    if args.list_only:
        print(projects)
        return 0

    if args.generate_registers_only:
        # Generate register output from all modules. Note that this is not used by the
        # build flow or simulation flow, it is only for the user to inspect the artifacts.
        generate_register_artifacts(
            modules=modules, output_path=args.projects_path.parent / "registers"
        )
        return 0

    if args.open:
        projects.open(projects_path=args.projects_path)
        return 0

    if args.collect_artifacts_only:
        # We have to assume that the project exists if the user sent this argument.
        # The 'collect_artifacts_function' call below will probably fail if it does not.
        create_ok = True

    elif args.use_existing_project:
        create_ok = projects.create_unless_exists(
            projects_path=args.projects_path,
            num_parallel_builds=args.num_parallel_builds,
            ip_cache_path=args.ip_cache_path,
        )

    else:
        create_ok = projects.create(
            projects_path=args.projects_path,
            num_parallel_builds=args.num_parallel_builds,
            ip_cache_path=args.ip_cache_path,
        )

    if not create_ok:
        return 1

    if args.create_only:
        return 0

    # If doing only synthesis, there are no artifacts to collect.
    collect_artifacts_function = (
        None if (args.synth_only or args.netlist_builds) else collect_artifacts_function
    )

    if args.collect_artifacts_only:
        assert collect_artifacts_function is not None, "No artifact collection available"

        for project in projects.projects:
            # Assign the arguments in the exact same way as the call to 'projects.build()' below.
            # Ensures that the correct output path is used in all scenarios.
            project_build_output_path = projects.get_build_project_output_path(
                project=project, projects_path=args.projects_path, output_path=args.output_path
            )
            # Collect artifacts function must return True.
            assert collect_artifacts_function(project, project_build_output_path)

        return 0

    build_ok = projects.build(
        projects_path=args.projects_path,
        num_parallel_builds=args.num_parallel_builds,
        num_threads_per_build=args.num_threads_per_build,
        output_path=args.output_path,
        collect_artifacts=collect_artifacts_function,
        synth_only=args.synth_only,
        from_impl=args.from_impl,
    )

    if build_ok:
        return 0

    return 1


def generate_register_artifacts(modules: "ModuleList", output_path: Path) -> None:
    """
    Generate register artifacts from the given modules.

    Arguments:
        modules: Registers from these modules will be included.
        output_path: Register artifacts will be placed here.
    """
    print(f"Generating register artifacts in {output_path.resolve()}...")

    # Empty the output directory so we don't have leftover old artifacts.
    create_directory(directory=output_path, empty=True)

    for module in modules:
        if module.registers is not None:
            CHeaderGenerator(module.registers, output_path / "c").create()

            CppInterfaceGenerator(module.registers, output_path / "cpp" / "include").create()
            CppHeaderGenerator(module.registers, output_path / "cpp" / "include").create()
            CppImplementationGenerator(module.registers, output_path / "cpp").create()

            HtmlPageGenerator(module.registers, output_path / "html").create()

            PythonPickleGenerator(module.registers, output_path / "python").create()



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from pathlib import Path
from typing import TYPE_CHECKING, Optional

# Third party libraries
from vunit.ui import VUnit

# Local folder libraries
from .simlib_commercial import VivadoSimlibCommercial
from .simlib_ghdl import VivadoSimlibGhdl

if TYPE_CHECKING:
    # Local folder libraries
    from .simlib_common import VivadoSimlibCommon


class VivadoSimlib:
    """
    Factory class for getting a Vivado simlib API.
    """

    @staticmethod
    def init(
        output_path: Path, vunit_proj: VUnit, vivado_path: Optional[Path] = None
    ) -> "VivadoSimlibCommon":
        """
        Get a Vivado simlib API suitable for your current simulator. Uses VUnit mechanism
        for detecting the simulator currently in use.

        Will return a :class:`.VivadoSimlibCommon` subclass object.

        Arguments:
            output_path: The compiled simlib will be placed here.
            vunit_proj: The VUnit project that is used to run simulation.
            vivado_path: Path to Vivado executable. If left out, the default
                from system ``PATH`` will be used.
        """
        simulator_interface = vunit_proj._simulator_class  # pylint: disable=protected-access

        if simulator_interface.name == "ghdl":
            return VivadoSimlibGhdl(
                vivado_path=vivado_path,
                output_path=output_path,
                vunit_proj=vunit_proj,
                simulator_interface=simulator_interface,
            )

        return VivadoSimlibCommercial(
            vivado_path=vivado_path,
            output_path=output_path,
            vunit_proj=vunit_proj,
            simulator_interface=simulator_interface,
        )



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------


REPOSITORY_URL = "https://github.com/tsfpga/tsfpga"
WEBSITE_URL = "https://tsfpga.com"


def get_short_slogan() -> str:
    """
    Short slogan used e.g. on pypi.org.
    Note that there seems to be an upper limit of 98 characters when rendering the slogan
    on pypi.org.

    Note that this slogan should be the same as the one used in the readme and on the website below.
    The difference is capitalization and whether the project name is included.
    """
    result = "A flexible and scalable development platform for modern FPGA projects"
    return result


def get_readme_rst(
    include_extra_for_github: bool = False,
    include_extra_for_website: bool = False,
    include_extra_for_pypi: bool = False,
) -> str:
    """
    Get the complete README.rst for tsfpga (to be used on website and in PyPI release).

    The arguments control some extra text that is included. This is mainly links to the
    other places where you can find information on the project (website, GitHub, PyPI).

    Arguments:
        include_extra_for_github (bool): Include the extra text that shall be included in the
            GitHub README.
        include_extra_for_website (bool): Include the extra text that shall be included in the
            website main page.
      include_extra_for_pypi (bool): Include the extra text that shall be included in the
            PyPI release README.
    """
    if include_extra_for_github:
        readme_rst = ""
        extra_rst = f"""\
**See documentation on the website**: {WEBSITE_URL}

**See PyPI for installation details**: https://pypi.org/project/tsfpga/
"""
    elif include_extra_for_website:
        # The website needs the initial heading, in order for the landing page to get
        # the correct title.
        # The others do not need this initial heading, it just makes the GitHub/PyPI page
        # more clunky.
        readme_rst = """\
About tsfpga
============

"""
        extra_rst = f"""\
To install the Python package, see :ref:`installation`.
To check out the source code go to the `GitHub page <{REPOSITORY_URL}>`__.
"""
    elif include_extra_for_pypi:
        readme_rst = ""
        extra_rst = f"""\
**See documentation on the website**: {WEBSITE_URL}

**Check out the source code on GitHub**: {REPOSITORY_URL}
"""
    else:
        readme_rst = ""
        extra_rst = ""

    readme_rst += f"""\
.. image:: {WEBSITE_URL}/logos/banner.png
  :alt: Project banner
  :align: center

|

.. |pic_website| image:: {WEBSITE_URL}/badges/website.svg
  :alt: Website
  :target: {WEBSITE_URL}

.. |pic_repository| image:: {WEBSITE_URL}/badges/repository.svg
  :alt: Repository
  :target: {REPOSITORY_URL}

.. |pic_chat| image:: {WEBSITE_URL}/badges/chat.svg
  :alt: Chat
  :target: {REPOSITORY_URL}/discussions

.. |pic_pip_install| image:: {WEBSITE_URL}/badges/pip_install.svg
  :alt: pypi
  :target: https://pypi.org/project/tsfpga/

.. |pic_license| image:: {WEBSITE_URL}/badges/license.svg
  :alt: License
  :target: {WEBSITE_URL}/license_information.html

.. |pic_ci_status| image:: {REPOSITORY_URL}/actions/workflows/ci.yml/\
badge.svg?branch=main
  :alt: CI status
  :target: {REPOSITORY_URL}/actions/workflows/ci.yml

.. |pic_python_line_coverage| image:: {WEBSITE_URL}/badges/python_coverage.svg
  :alt: Python line coverage
  :target: {WEBSITE_URL}/python_coverage_html

|pic_website| |pic_repository| |pic_pip_install| |pic_license| |pic_chat| |pic_ci_status|
|pic_python_line_coverage|

tsfpga is a flexible and scalable development platform for modern FPGA projects.
With its Python-based build/simulation flow it is perfect for CI/CD and test-driven development.
The API is simple and easy to use
(a complete `simulation project <{WEBSITE_URL}/simulation.html>`__ is set up in less than
15 lines).

{extra_rst}
Key features
------------

* Source code-centric `project structure <{WEBSITE_URL}/module_structure.html>`__
  for scalability.
  Build projects, test configurations, constraints, IP cores, etc. are handled close to the
  source code, not in a central monolithic script.
* Automatically adds build/simulation sources if a recognized folder structure is used.
* Enables `local VUnit test configuration
  <{WEBSITE_URL}/simulation.html#local-configuration-of-test-cases>`__ without
  multiple ``run.py``.
* Handling of `IP cores <{WEBSITE_URL}/simulation.html#simulating-with-vivado-ip-cores>`__
  and `simlib <{WEBSITE_URL}/simulation.html#vivado-simulation-libraries>`__
  for your simulation project, with automatic re-compile when needed.
* Python-based `Vivado build system <{WEBSITE_URL}/fpga_build.html>`__ where many builds can
  be run in parallel.
* Tightly integrated with `hdl-registers <https://hdl-registers.com>`__.
  Register code generation is performed before each simulation and each build.
* Released under the very permissive BSD 3-Clause License.

The maintainers place high focus on quality, with everything having good unit test coverage and a
thought-out structure.
The project is mature and used in many production environments.
"""

    return readme_rst



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import subprocess
from pathlib import Path

# Third party libraries
import pytest

# First party libraries
from tsfpga.system_utils import (
    create_directory,
    create_file,
    delete,
    file_is_in_directory,
    path_relative_to,
    read_file,
    read_last_lines_of_file,
    run_command,
)


def test_delete_files_and_folders(tmp_path):
    # Test deleting a file, with and without wait
    path = create_file(tmp_path / "temp.txt")
    assert path.exists()
    delete(path)
    assert not path.exists()

    path = create_file(tmp_path / "temp.txt")
    assert path.exists()
    delete(path, wait_until_deleted=True)
    assert not path.exists()

    # Test deleting a directory, with and without wait
    path = create_directory(tmp_path / "temp_dir")
    assert path.exists()
    delete(path)
    assert not path.exists()

    path = create_directory(tmp_path / "temp_dir")
    assert path.exists()
    delete(path, wait_until_deleted=True)
    assert not path.exists()


def test_create_directory_plain(tmp_path):
    path = tmp_path / "temp_dir"
    assert not path.exists()

    create_directory(path)
    assert path.exists()
    assert path.is_dir()


def test_create_directory_that_exists_without_empty(tmp_path):
    path = tmp_path / "temp_dir"
    sub_path = create_directory(path / "sub")

    create_directory(path, empty=False)
    assert sub_path.exists()


def test_create_directory_that_exists_with_empty(tmp_path):
    path = tmp_path / "temp_dir"
    sub_path = create_directory(path / "sub")

    create_directory(path)
    assert path.exists()
    assert not sub_path.exists()


def test_create_directory_without_empty_when_path_is_a_file(tmp_path):
    path = create_file(tmp_path / "file.txt", contents="data")

    with pytest.raises(FileExistsError) as exception_info:
        create_directory(path, empty=False)
    assert str(exception_info.value) == f"Requested directory path already exists as a file: {path}"

    assert read_file(path) == "data"


def test_file_is_in_directory(tmp_path):
    assert file_is_in_directory(tmp_path / "file.txt", [tmp_path])
    assert not file_is_in_directory(tmp_path / "file.txt", [tmp_path / "sub"])

    assert file_is_in_directory(
        tmp_path / "sub" / "file.txt", [tmp_path / "sub", tmp_path / "sub2"]
    )
    assert not file_is_in_directory(tmp_path / "file.txt", [tmp_path / "sub", tmp_path / "sub2"])
    assert not file_is_in_directory(
        tmp_path / "sub" / "file.txt", [tmp_path / "sub2", tmp_path / "sub3"]
    )


def test_path_relative_to():
    this_file = Path(__file__)
    this_dir = this_file.parent
    parent = this_dir.parent

    assert path_relative_to(this_file, this_dir) == Path(this_file.name)
    assert path_relative_to(this_file, parent) == Path(this_dir.name) / this_file.name
    assert (
        path_relative_to(this_file, parent / "whatever")
        == Path("..") / this_dir.name / this_file.name
    )


def test_read_last_lines_of_file_with_short_file(tmp_path):
    # A file that is smaller than the buffer size
    data = "a\nb\nc"
    file = create_file(tmp_path / "data.txt", contents=data)
    assert read_last_lines_of_file(file, num_lines=10) == data


def test_read_last_lines_of_file_with_long_file(tmp_path):
    # A file that is larger than the buffer size
    data_trash = (("a" * 700) + "\n") * 3000
    data_last_lines = (("b" * 700) + "\n") * 10
    file = create_file(tmp_path / "data.txt", contents=data_trash + data_last_lines)
    assert read_last_lines_of_file(file, num_lines=10) == data_last_lines


def test_read_last_lines_of_file_with_trailing_newlines(tmp_path):
    # A file that is smaller than the buffer size
    data = "a\nb\n\n  \n\n"
    file = create_file(tmp_path / "data.txt", contents=data)
    assert read_last_lines_of_file(file, num_lines=10) == data


def test_read_last_lines_of_file_with_empty_file(tmp_path):
    data = ""
    file = create_file(tmp_path / "data.txt", contents=data)
    assert read_last_lines_of_file(file, num_lines=10) == data

    data = "\n"
    file = create_file(tmp_path / "data.txt", contents=data)
    assert read_last_lines_of_file(file, num_lines=10) == data


def test_run_command_called_with_nonexisting_binary_should_raise_exception():
    cmd = ["/apa/hest/zebra.exe", "foobar"]
    with pytest.raises(FileNotFoundError):
        run_command(cmd)


def test_run_command_with_non_zero_return_code_should_raise_exception():
    cmd = ["ls", "/apa/hest/zebra"]
    with pytest.raises(subprocess.CalledProcessError):
        run_command(cmd)


def test_run_command_called_with_non_list_should_raise_exception():
    cmd = ["ls", "-la"]
    run_command(cmd)

    cmd = "ls -la"
    with pytest.raises(ValueError) as exception_info:
        run_command(cmd)
    assert str(exception_info.value).startswith("Must be called with a list")


def test_run_command_should_capture_output_as_strings():
    this_dir = Path(__file__).parent.resolve()

    cmd = ["ls", str(this_dir)]
    result = run_command(cmd, capture_output=True)

    assert isinstance(result.stdout, str)
    assert isinstance(result.stderr, str)

    assert result.stderr == ""

    print(result.stdout)
    print(result.stderr)

    # Show that it is regular text with regular newlines.
    assert "\ntest_system_utils.py\n" in result.stdout
    assert "\ntest_ip_core_file.py\n" in result.stdout



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from unittest import mock

# First party libraries
from tsfpga.svn_utils import find_svn_files, svn_local_changes_are_present
from tsfpga.system_utils import create_file

# SVN is pretty much impossible to work with. Most of the functions are untested.


@mock.patch("tsfpga.svn_utils.run_command", autospec=True)
def test_svn_local_changes_are_present(run_command):
    run_command.return_value.stdout = """
?       .vscode
?       build
"""
    assert not svn_local_changes_are_present()

    run_command.return_value.stdout = """
?       .vscode
M       build.py
?       build
"""
    assert svn_local_changes_are_present()

    run_command.return_value.stdout = """
?       .vscode
!       build.py
?       build
"""
    assert svn_local_changes_are_present()


@mock.patch("tsfpga.svn_utils.run_command", autospec=True)
def test_find_svn_files(run_command, tmp_path):
    apa_txt = create_file(tmp_path / "apa.txt")
    hest_txt = create_file(tmp_path / "stuff" / "hest.txt")
    zebra_vhd = create_file(tmp_path / "stuff" / "zebra.vhd")
    zebra_pdf = create_file(tmp_path / "things" / "zebra.pdf")

    run_command.return_value.stdout = """
            104134   104134 lukas.vik    .
?                                        .pytest_cache
?                                        .vscode
?                                        build
            104134   104134 lukas.vik    apa.txt

Performing status on external item at 'stuff':
            104134   103740 lukas.vik    stuff
M           104134   103740 lukas.vik    stuff/hest.txt
            104134   103740 lukas.vik    stuff/zebra.vhd

Performing status on external item at 'things':
            104134   103740 lukas.vik    things
M   X       104134   103740 lukas.vik    things/zebra.pdf
"""
    assert set(find_svn_files(tmp_path)) == {apa_txt, hest_txt, zebra_pdf, zebra_vhd}

    # Filter file endings include
    assert set(find_svn_files(tmp_path, file_endings_include="txt")) == {apa_txt, hest_txt}
    assert set(find_svn_files(tmp_path, file_endings_include=("txt", "vhd"))) == {
        apa_txt,
        hest_txt,
        zebra_vhd,
    }

    # Filter file endings avoid
    assert set(find_svn_files(tmp_path, file_endings_avoid="txt")) == {zebra_vhd, zebra_pdf}
    assert set(find_svn_files(tmp_path, file_endings_avoid=("txt", "vhd"))) == {zebra_pdf}

    # Filter exclude
    assert set(find_svn_files(tmp_path, excludes=[zebra_vhd.parent])) == {apa_txt, zebra_pdf}
    assert set(find_svn_files(tmp_path, excludes=[zebra_vhd.parent, zebra_pdf.parent])) == {apa_txt}



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import copy
from typing import TYPE_CHECKING, Iterator

if TYPE_CHECKING:
    # Local folder libraries
    from .module import BaseModule


class ModuleList:
    """
    Wrapper for a list of modules, with convenience functions.
    """

    def __init__(self) -> None:
        self._modules: list["BaseModule"] = []

    def append(self, module: "BaseModule") -> None:
        """
        Append a module to the list.
        """
        self._modules.append(module)

    def get(self, module_name: "str") -> "BaseModule":
        """
        Get the module with the specified name. If no module matched, an exception is raised.
        """
        for module in self._modules:
            if module.name == module_name:
                return module

        raise ValueError(f'No module "{module_name}" available')

    def __iter__(self) -> Iterator["BaseModule"]:
        return iter(self._modules)

    def __getitem__(self, index: int) -> "BaseModule":
        return self._modules[index]

    def __len__(self) -> int:
        return len(self._modules)

    def __add__(self, other: "ModuleList") -> "ModuleList":
        if not isinstance(other, self.__class__):
            raise TypeError(f"Can only concatenate with another {self.__class__.__name__}")

        # Note that the list concatenation implies a shallow copy of the lists
        result = self.__class__()
        result._modules = self._modules + other._modules
        return result

    def __copy__(self) -> "ModuleList":
        result = self.__class__()
        result._modules = self._modules.copy()
        return result

    def copy(self) -> "ModuleList":
        """
        Create a shallow copy of the module list. This public function is available as a
        convenience and to mimic the interface of a regular python list.
        """
        return copy.copy(self)

    def __str__(self) -> str:
        return str(self._modules)



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from unittest.mock import MagicMock

# Third party libraries
import pytest

# First party libraries
from tsfpga.module_list import ModuleList


def test_module_list_indexing():
    modules = ModuleList()

    module_a = MagicMock()
    module_b = MagicMock()

    modules.append(module_a)
    modules.append(module_b)

    assert modules[0] is module_a
    assert modules[1] is module_b


@pytest.fixture
def module_list_test():
    class ModuleListTest:
        def __init__(self):
            self.modules = ModuleList()

            self.module_a = MagicMock()
            self.module_b = MagicMock()
            self.module_c = MagicMock()

            self.module_a.name = "a"
            self.module_b.name = "b"
            self.module_c.name = "c"

            self.modules.append(self.module_a)
            self.modules.append(self.module_b)
            self.modules.append(self.module_c)

    return ModuleListTest()


# False positive for pytest fixtures
# pylint: disable=redefined-outer-name


def test_get(module_list_test):
    assert module_list_test.modules.get("a") == module_list_test.module_a
    assert module_list_test.modules.get("c") == module_list_test.module_c

    with pytest.raises(ValueError) as exception_info:
        module_list_test.modules.get("d")
    assert str(exception_info.value) == 'No module "d" available'


def test_iteration(module_list_test):
    expected = [module_list_test.module_a, module_list_test.module_b, module_list_test.module_c]

    # pylint: disable=unnecessary-comprehension
    assert [module for module in module_list_test.modules] == expected
    assert list(module_list_test.modules) == expected


def test_can_cast_to_string_without_error(module_list_test):
    print(module_list_test.modules)


def test_concatenation(module_list_test):
    modules_2 = ModuleList()

    module_d = MagicMock()
    module_d.name = "d"
    modules_2.append(module_d)

    all_modules = module_list_test.modules + modules_2

    assert len(all_modules) == 4
    assert list(all_modules) == [
        module_list_test.module_a,
        module_list_test.module_b,
        module_list_test.module_c,
        module_d,
    ]

    # The addition should result in a new object. Should not affect the existing objects.
    assert len(module_list_test.modules) == 3
    assert len(modules_2) == 1

    assert (all_modules is not modules_2) and (all_modules is not module_list_test.modules)

    # pylint: disable=protected-access
    assert (all_modules._modules is not modules_2._modules) and (
        all_modules._modules is not module_list_test.modules._modules
    )


def test_concatenation_with_unknown_object_should_raise_interrupt(module_list_test):
    with pytest.raises(TypeError) as exception_info:
        # pylint: disable=pointless-statement
        module_list_test.modules + 3
    assert str(exception_info.value) == "Can only concatenate with another ModuleList"


def test_copy(module_list_test):
    modules_copy = module_list_test.modules.copy()

    modules_copy.append(4)
    assert len(module_list_test.modules) == 3
